
# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2006, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import errno
import math
import sys
import random
import time

try:
    import sqlite3 as sqlite
except ImportError:
    from pysqlite2 import dbapi2 as sqlite

import events, hub, requests
import errors
import log
import metadata
import item
import service
import encoding


create_tables = """
CREATE TABLE artists (
  id             INTEGER CONSTRAINT pk_artist_id PRIMARY KEY AUTOINCREMENT,
  name           TEXT UNIQUE
);

CREATE TABLE albums (
  id             INTEGER CONSTRAINT pk_album_id PRIMARY KEY AUTOINCREMENT,
  artist_id      INTEGER CONSTRAINT fk_albums_artist_id REFERENCES artists(id),
  name           TEXT,
  UNIQUE (artist_id, name)
);

CREATE TABLE tags (
  id             INTEGER CONSTRAINT pk_tag_id PRIMARY KEY AUTOINCREMENT,
  name           TEXT UNIQUE
);

CREATE TABLE taggings (
  song_id        INTEGER CONSTRAINT fk_song_id REFERENCES songs(id),
  tag_id         INTEGER CONSTRAINT fk_tag_id  REFERENCES tags(id)
);

CREATE TABLE playstats (
  song_id        INTEGER CONSTRAINT fk_song_id REFERENCES songs(id),
  date_played    TIMESTAMP
);

CREATE TABLE playlists (
  id             INTEGER CONSTRAINT pk_playlist_id PRIMARY KEY AUTOINCREMENT,
  name           TEXT UNIQUE
);

CREATE TABLE playlistcontents (
  playlist_id    INTEGER CONSTRAINT fk_playlist_id  REFERENCES playlists(id),
  song_id        INTEGER CONSTRAINT fk_song_id      REFERENCES songs(id),
  position       INTEGER
);

CREATE TABLE songs (
  id                    INTEGER CONSTRAINT pk_song_id PRIMARY KEY AUTOINCREMENT,
  url                   TEXT UNIQUE,
  type                  TEXT,
  title                 TEXT,
  album_id              INTEGER CONSTRAINT fk_song_album_id  REFERENCES albums(id),
  artist_id             INTEGER CONSTRAINT fk_song_artist_id REFERENCES artists(id),
  album_artist_id       INTEGER CONSTRAINT fk_song_artist_id REFERENCES artists(id),
  year                  INTEGER,
  comments              BLOB,
  lyrics                BLOB,
  bpm                   INTEGER,
  length                INTEGER,
  tracknumber           INTEGER,
  trackcount            INTEGER,
  disknumber            INTEGER,
  diskcount             INTEGER,
  compilation           BOOL,
  bitrate               INTEGER,
  is_vbr                BOOL,
  samplerate            INTEGER,
  replaygain_track_gain FLOAT,
  replaygain_track_peak FLOAT,
  replaygain_album_gain FLOAT,
  replaygain_album_peak FLOAT,
  size                  INTEGER,
  date_added            TIMESTAMP,
  date_updated          TIMESTAMP,
  date_lastplayed       TIMESTAMP,
  playcount             INTEGER,
  skipcount             INTEGER,
  rating                INTEGER
);

CREATE INDEX album_id ON albums(name);
CREATE INDEX artist_id ON artists(name);
CREATE INDEX tag_id ON tags(name);
CREATE INDEX playlist_id ON playlists(name);

CREATE INDEX url_song ON songs(url);
CREATE INDEX album_id_song ON songs(album_id);
CREATE INDEX artist_id_song ON songs(artist_id);
CREATE INDEX year_song ON songs(year);
CREATE INDEX compilation_song ON songs(compilation);

CREATE INDEX taggings_song_id ON taggings(song_id);
CREATE INDEX taggings_tag_id ON taggings(tag_id);

CREATE INDEX playlistcontents_song_id ON playlistcontents(song_id);
CREATE INDEX playlistcontents_playlist_id ON playlistcontents(playlist_id);
"""

songcolumns_plain = ["url", "type", "title",  "year", "bpm",
                     "length", "tracknumber", "trackcount", "disknumber", "diskcount",
                     "compilation", "bitrate", "is_vbr", "samplerate", 
                     "replaygain_track_gain", "replaygain_track_peak",
                     "replaygain_album_gain", "replaygain_album_peak", 
                     "size", "compilation", "date_added", "date_updated", "date_lastplayed",
                     "playcount", "skipcount", "rating"]

songcolumns_indices = ["album_id", "artist_id", "album_artist_id"]
songcolumns_w_indices = songcolumns_plain + songcolumns_indices
songcolumns_lists = ["comments", "lyrics"]
songcolumns_all = songcolumns_w_indices + songcolumns_lists

# secure unpickler which does not accept any instances

import cPickle, cStringIO

def loads(s):
    unpickler = cPickle.Unpickler(cStringIO.StringIO(s))
    unpickler.find_global = None
    return unpickler.load()

def dumps(obj):
    return buffer(cPickle.dumps(obj))

#
# statistical information about songdb
#

class songdbstats:
    def __init__(self, id, type, basedir, location, dbfile, cachesize,
                 numberofsongs, numberofalbums, numberofartists, numberoftags):
        self.id = id
        self.type = type
        self.basedir = basedir
        self.location = location
        self.dbfile = dbfile
        self.cachesize = cachesize
        self.numberofsongs = numberofsongs
        self.numberofalbums = numberofalbums
        self.numberofartists = numberofartists
        self.numberoftags = numberoftags

#
# songdb class
#

class songdb(service.service):

    currentdbversion = 1

    def __init__(self, id, config, songdbhub):
        service.service.__init__(self, "%r songdb" % id, hub=songdbhub)
        self.id = id
        self.basedir = config.musicbasedir
        self.dbfile = config.dbfile
        self.cachesize = config.cachesize
        self.playingstatslength = config.playingstatslength

        if not os.path.isdir(self.basedir):
            raise errors.configurationerror("musicbasedir '%r' of database %r is not a directory." % 
                                            (self.basedir, self.id))

        if not os.access(self.basedir, os.X_OK | os.R_OK):
            raise errors.configurationerror("you are not allowed to access and read config.general.musicbasedir.")

        # currently active cursor - initially, none
        self.cur = None

        # we need to be informed about database changes
        self.channel.subscribe(events.add_song, self.add_song)
        self.channel.subscribe(events.update_song, self.update_song)
        self.channel.subscribe(events.delete_song, self.delete_song)
        self.channel.subscribe(events.song_played, self.song_played)
        self.channel.subscribe(events.song_skipped, self.song_skipped)

        self.channel.subscribe(events.add_playlist, self.add_playlist)
        self.channel.subscribe(events.update_playlist, self.update_playlist)
        self.channel.subscribe(events.delete_playlist, self.delete_playlist)

        self.channel.subscribe(events.clearstats, self.clearstats)

        # we are a database service provider...
        self.channel.supply(requests.getdatabasestats, self.getdatabasestats)
        self.channel.supply(requests.getsong_metadata, self.getsong_metadata)
        self.channel.supply(requests.getartists, self.getartists)
        self.channel.supply(requests.getalbums, self.getalbums)
        self.channel.supply(requests.gettag_id, self.gettag_id)
        self.channel.supply(requests.getsongs, self.getsongs)
        self.channel.supply(requests.getnumberofsongs, self.getnumberofsongs)
        self.channel.supply(requests.getnumberofalbums, self.getnumberofalbums)
        self.channel.supply(requests.getnumberofartists, self.getnumberofartists)
        self.channel.supply(requests.getnumberoftags, self.getnumberoftags)
        self.channel.supply(requests.getnumberofratings, self.getnumberofratings)
        self.channel.supply(requests.gettags, self.gettags)
        self.channel.supply(requests.getratings, self.getratings)
        self.channel.supply(requests.getlastplayedsongs, self.getlastplayedsongs)
        self.channel.supply(requests.getplaylists, self.getplaylists)

        self.autoregisterer = songautoregisterer(self.basedir, self.id, self.isbusy,
                                                 config.tracknrandtitlere, config.postprocessors)
        self.autoregisterer.start()

    def run(self):
        # self.con = sqlite.connect(":memory:")
        log.debug("dbfile: '%r'" % self.dbfile)
        self.con = sqlite.connect(self.dbfile)
        self.con.row_factory = sqlite.Row

        dbversion = self.con.execute("PRAGMA user_version").fetchone()[0]
        log.debug("Found on-disk db version: %d" % dbversion)
        if dbversion == 0:
            # fresh database
            self._txn_begin()
            self.con.executescript(create_tables)
            self._txn_commit()
            self.con.execute("PRAGMA user_version=%d" % self.currentdbversion)
        log.debug("Starting db sevice")
        service.service.run(self)
        self.close()

    def close(self):
        self.con.close()

    # transaction machinery

    def _txn_begin(self):
        if self.cur:
            raise RuntimeError("more than one transaction in parallel is not supported")
        # self.con.execute("BEGIN TRANSACTION")
        self.cur = self.con.cursor()

    def _txn_commit(self):
        # self.con.execute("COMMIT TRANSACTION")
        self.cur.close()
        self.con.commit()
        self.cur = None

    def _txn_abort(self):
        # self.con.execute("ROLLBACK")
        self.con.rollback()
        self.cur.close()
        self.cur = None

    # resetting db stats

    def _clearstats(self):
        self._txn_begin()
        try:
            self.cur.execute("DELETE FROM playstats")
            self.cur.execute("UPDATE songs SET playcount = 0, skipcount = 0, date_lastplayed = NULL ")

        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()

    #
    # methods for adding, updating and deleting songs
    #

    # helper methods

    def _queryindex(self, table, indexnames, values):
        " query indexnames in table and return id "
        newindexentry = False
        wheres = " AND ".join(["%s = ?" % indexname for indexname in indexnames])
        self.cur.execute("SELECT id FROM %s WHERE %s" % (table, wheres), values)
        r = self.cur.fetchone()
        return r["id"]

    def _queryregisterindex(self, table, indexnames, values):
        " register in table and return if tuple (id, newentry) "
        newindexentry = False
        wheres = " AND ".join(["%s = ?" % indexname for indexname in indexnames])
        self.cur.execute("SELECT id FROM %s WHERE %s" % (table, wheres), values)
        r = self.cur.fetchone()
        if r is None:
            self.cur.execute("INSERT INTO %s (%s) VALUES (%s)" % (table, ", ".join(indexnames),
                                                                  ", ".join(["?"]*len(indexnames))), 
                             values)
            self.cur.execute("SELECT id FROM %s WHERE %s" % (table, wheres), values)
            r = self.cur.fetchone()
            newindexentry = True
        return r["id"], newindexentry

    def _checkremoveindex(self, indextable, reftable, indexnames, value):
        "remove entry from indextable if no longer referenced in reftable and return whether this has happened"
        if value is None:
            return False
        wheres = " OR ".join(["%s = ?" % indexname for indexname in indexnames])
        num = self.cur.execute("SELECT count(*) FROM %s WHERE (%s)" % (reftable, wheres),
                               [value]*len(indexnames)).fetchone()[0]
        if num == 0:
            self.cur.execute("DELETE FROM %s WHERE id = ?" % indextable, [value])
            return True
        else:
            return False

    _song_insert = "INSERT INTO songs (%s) VALUES (%s)" % (",".join(songcolumns_all),
                                                           ",".join(["?"] * len(songcolumns_all)))

    def _add_song(self, song):
        """add song metadata to database"""
        log.debug("adding song: %r" % song)

        if not isinstance(song, metadata.song_metadata):
            log.error("add_song: song has to be a meta.song instance, not a %r instance" % 
                      song.__class__)
            return

        self._txn_begin()
        try:
            # query and register artist, album_artist and album
            if song.artist:
                song.artist_id, newartist = self._queryregisterindex("artists", ["name"], [song.artist])
            else:
                song.artist_id, newartist = None, False
            if song.album_artist:
                song.album_artist_id, newartist2 = self._queryregisterindex("artists", ["name"], 
                                                                            [song.album_artist])
                newartist = newartist or newartist2
                if song.album:
                    song.album_id, newalbum = self._queryregisterindex("albums", ["artist_id", "name"], 
                                                                       [song.album_artist_id, song.album])
                else:
                    song.album_id, newalbum = None, False
            else:
                song.album_artist_id = None
                song.album_id = None
                newalbum = False

            # pickle the comments and lyrics lists
            comments = dumps(song.comments)
            lyrics = dumps(song.lyrics)

            # register song
            self.cur.execute(self._song_insert,
                             [getattr(song, columnname) for columnname in songcolumns_w_indices] + [comments, lyrics])

            self.cur.execute("SELECT id FROM songs WHERE url = ?", (song.url,))
            r = self.cur.fetchone()
            song_id = r["id"]

            # register song tags
            newtag = False
            for tag in song.tags:
                tag_id, newtag2 = self._queryregisterindex("tags", ["name"], [tag])
                newtag = newtag or newtag2
                self.cur.execute("INSERT INTO taggings (song_id, tag_id) VALUES (?, ?)", 
                                 (song_id, tag_id))
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()
            if newartist:
                hub.notify(events.artistschanged(self.id))
            if newalbum:
                hub.notify(events.albumschanged(self.id))
            if newtag:
                hub.notify(events.tagschanged(self.id))
            # we don't issue a songschanged event because the resulting queries put a too high load 
            # on the database
            # hub.notify(events.songschanged(self.id))

            #for r in cur.execute("SELECT id, name FROM artists"):
            #    log.info("AR: %r %r" % (r["id"], r["name"]))
            #for r in cur.execute("SELECT id, artist_id, name FROM albums"):
            #    log.info("AL: %r %r %r" % (r["id"], r["artist_id"], r["name"]))
            #for r in cur.execute("SELECT id, title FROM songs"):
            #    log.info("S: %r %r" % (r["id"], r["title"]))

    def _delete_song(self, song):
        """delete song from database"""
        log.debug("delete song: %r" % song)
        if not isinstance(song, item.song):
            log.error("_delete_song: song has to be a item.song instance, not a %r instance" % song.__class__)

        self._txn_begin()
        try:
            # remove song
            self.cur.execute("DELETE FROM songs WHERE id = ?", [song.id])

            # remove corresponding album and artists
            deletedalbum = self._checkremoveindex("albums", "songs", ["album_id"], song.album_id)
            deletedartist = self._checkremoveindex("artists", "songs", ["album_artist_id", "artist_id"], 
                                                   song.artist_id)
            deletedartist |= self._checkremoveindex("artists", "songs", ["album_artist_id", "artist_id"], 
                                                    song.album_artist_id)


            # query tags in order to be able to delete them (as opposed to album_id, etc.,
            # they are not stored in item.song)
            tag_ids = []
            for r in self.cur.execute("""SELECT DISTINCT tags.id AS tag_id FROM tags
                                         JOIN taggings ON (taggings.tag_id =tags.id)
                                         WHERE taggings.song_id = ?""", [song.id]):
                tag_ids.append(r["tag_id"])

            # remove taggings
            deletedtag = False
            self.cur.execute("DELETE FROM taggings WHERE song_id = ?", [song.id])
            for tag_id in tag_ids:
                deletedtag |= self._checkremoveindex("tags", "taggings", ["tag_id"], tag_id)
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()
            if deletedartist:
                hub.notify(events.artistschanged(self.id))
            if deletedalbum:
                hub.notify(events.albumschanged(self.id))
            if deletedtag:
                hub.notify(events.tagschanged(self.id))
        # XXX send event?

    _song_update = ( "INSERT OR REPLACE INTO songs (id, %s) VALUES (?, %s)" % 
                     (",".join(songcolumns_all), ",".join(["?"] * len(songcolumns_all))) )

    def _update_song(self, song):
        """updates entry of song"""
        log.debug("updating song %r" % song)
        if not isinstance(song, item.song):
            log.error("_update_song: song has to be a item.song instance, not a %r instance" %
                      newsong.__class__)
            return
        if not song.song_metadata:
            log.error("_update_song: song doesn't contain song metadata")
            return
        oldsong = self._getsong_metadata(song.id)

        self._txn_begin()
        try:

            # query artist, album_artist and album of oldsong (in comparison to the _add_song method, we do not have
            # to add entries to the indices
            if oldsong.artist:
                oldsong.artist_id = self._queryindex("artists", ["name"], [oldsong.artist])
            else:
                oldsong.artist_id = None
            if oldsong.album_artist:
                oldsong.album_artist_id = self._queryindex("artists", ["name"], [oldsong.album_artist])
                if oldsong.album:
                    oldsong.album_id = self._queryindex("albums", ["artist_id", "name"], [oldsong.album_artist_id, oldsong.album])
                else:
                    oldsong.album_id = None
            else:
                oldsong.album_artist_id = None
                oldsong.album_id = None

            # flags for changes of corresponding tables
            changedartists = False
            changedalbums = False
            changedtags = False
            # register new artists, album_artists and albums if necessary
            if oldsong.artist != song.artist:
                if song.artist:
                    song.artist_id, newartist = self._queryregisterindex("artists", ["name"], [song.artist])
                    changedartists |= newartist
                else:
                    song.artist_id = None
            if oldsong.album_artist != song.album_artist:
                if song.album_artist:
                    song.album_artist_id, newartist = self._queryregisterindex("artists", ["name"], [song.album_artist])
                    changedartists |= newartist
                    if song.album:
                        song.album_id, newalbum = self._queryregisterindex("albums", ["artist_id", "name"], 
                                                                           [song.album_artist_id, song.album])
                        changedalbums |= newalbum
                    else:
                        song.album_id = None
                else:
                    song.album_artist_id = None
                    song.album_id = None
            elif oldsong.album != song.album and song.album:
                # only the album name changed
                song.album_id, newalbum = self._queryregisterindex("albums", ["artist_id", "name"], 
                                                                   [song.album_artist_id, song.album])
                changedalbums |= newalbum

            # encode the comments and lyrics lists
            comments = dumps(song.comments)
            lyrics = dumps(song.lyrics)

            # update songs table
            self.cur.execute(self._song_update, 
                            [song.id]+[getattr(song, columnname) for columnname in songcolumns_w_indices] + [comments, lyrics])

            # delete old artists, album_artists and albums if necessary
            # we have to do this after the songs table has been updated, otherwise we
            # cannot detect whether we have to remove an album/artist or not
            if oldsong.album != song.album:
                changedalbums |= self._checkremoveindex("albums", "songs", ["album_id"], song.album_id)
            if oldsong.artist != song.artist:
                changedartists |= self._checkremoveindex("artists", "songs", ["album_artist_id", "artist_id"],
                                                         oldsong.artist_id)
            if oldsong.album_artist != song.album_artist:
                changedartists |= self._checkremoveindex("artists", "songs", ["album_artist_id", "artist_id"],
                                                         oldsong.album_artist_id)

            # update tag information if necessary
            if oldsong.tags != song.tags:
                # check for new tags
                for tag in song.tags:
                    if tag not in oldsong.tags:
                        tag_id, newtag = self._queryregisterindex("tags", ["name"], [tag])
                        changedtags |= newtag
                        self.cur.execute("INSERT INTO taggings (song_id, tag_id) VALUES (?, ?)", 
                                         (song.id, tag_id))
                # check for removed tags
                for tag in oldsong.tags:
                    if tag not in song.tags:
                        tag_id = self._queryregisterindex("tags", ["name"], [tag])[0]
                        self.cur.execute("DELETE FROM taggings WHERE (tag_id = ? AND song_id = ?)", [tag_id, song.id])
                        changedtags |= self._checkremoveindex("tags", "taggings", ["tag_id"], tag_id)
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()
            if changedartists:
                hub.notify(events.artistschanged(self.id))
            if changedalbums:
                hub.notify(events.albumschanged(self.id))
            if changedtags:
                hub.notify(events.tagschanged(self.id))
        hub.notify(events.songchanged(self.id, song))

    def _song_played(self, song, date_played):
        """register playing of song"""
        log.debug("playing song: %r" % song)
        if not isinstance(song, item.song):
            log.error("_update_song: song has to be an item.song instance, not a %r instance" % song.__class__)
            return
        self._txn_begin()
        try:
            self.cur.execute("INSERT INTO playstats (song_id, date_played) VALUES (?, ?)", [song.id, date_played])
            self.cur.execute("UPDATE songs SET playcount = playcount+1, date_lastplayed = ? WHERE id = ?", [date_played, song.id])
            song.playcount += 1
            song.date_lastplayed = date_played
            song.dates_played.append(date_played)
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()
        hub.notify(events.songchanged(self.id, song))

    def _song_skipped(self, song):
        """register skipping of song"""
        log.debug("skipping song: %r" % song)
        if not isinstance(song, item.song):
            log.error("_update_song: song has to be an item.song instance, not a %r instance" % song.__class__)
            return
        self._txn_begin()
        try:
            self.cur.execute("UPDATE songs SET skipcount = skipcount+1 WHERE id = ?", [song.id])
            song.skipcount += 1
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()
        hub.notify(events.songchanged(self.id, song))

    def _add_playlist(self, name, songs):
        log.debug("adding playlist %r" % name)
        if not songs:
            log.error("_add_playlist: cannot add empty playlist")

        self._txn_begin()
        try:
            self.cur.execute("INSERT INTO playlists (name) VALUES (?)", [name])
            self.cur.execute("SELECT id FROM playlists WHERE name = ?", [name])
            r = self.cur.fetchone()
            playlist_id = r["id"]
            for position, song in enumerate(songs):
                self.cur.execute("INSERT INTO playlistcontents (playlist_id, song_id, position) VALUES (?, ?, ?)",
                                 [playlist_id, song.id, position + 1])
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()

    def _delete_playlist(self, playlist):
        """delete playlist from database"""
        if not self.playlists.has_key(playlist.id):
            raise KeyError

        log.debug("delete playlist: %r" % playlist)
        self._txn_begin()
        try:
            self.playlists.delete(playlist.id, txn=self.cur)
            hub.notify(events.dbplaylistchanged(self.id, playlist))
        except:
            self._txn_abort()
            raise
        else:
            self._txn_commit()

    _update_playlist = _add_playlist

    # read-only methods for accesing the database

    ##########################################################################################
    # !!! It is not save to call any of the following methods when a transaction is active !!!
    ##########################################################################################

    _song_select = """SELECT %s, artists.name AS artist, albums.name AS album 
                      FROM songs 
                      LEFT JOIN albums ON albums.id == album_id
                      LEFT JOIN artists ON artists.id == songs.artist_id
                      WHERE songs.id = ?
                      """ % ", ".join([c for c in songcolumns_all if c!="artist_id"])

    _song_tags_select = """SELECT tags.name AS name FROM tags
                           JOIN taggings ON taggings.tag_id = tags.id
                           WHERE taggings.song_id = ?"""

    _song_playstats_select = "SELECT date_played FROM playstats WHERE song_id = ?"

    def _getsong_metadata(self, song_id):
        """return song entry with given song_id"""
        log.debug("Querying song metadata for id=%r" % song_id)
        try:
            r = self.con.execute(self._song_select, [song_id]).fetchone()
            if r:
                # fetch album artist
                if r["album_artist_id"] is not None:
                    select = """SELECT name FROM artists WHERE id = ?"""
                    album_artist = self.con.execute(select, (r["album_artist_id"],)).fetchone()["name"]
                else:
                    album_artist = None

                # fetch tags
                tags = []
                for tr in self.con.execute(self._song_tags_select, [song_id]):
                    tags.append(tr["name"])

                # fetch playstats
                dates_played = []
                for tr in self.con.execute(self._song_playstats_select, [song_id]):
                    dates_played.append(tr["date_played"])

                # generate and populate metadata
                md = metadata.song_metadata()
                for field in songcolumns_plain:
                    md[field] = r[field]
                md.album = r["album"]
                md.artist = r["artist"]
                md.album_artist = album_artist
                md.tags = tags
                md.comments = loads(r["comments"])
                md.lyrics = loads(r["lyrics"])
                md.dates_played = dates_played
                return md
            else:
                log.debug("Song '%d' not found in database" % song_id)
                return None
        except:
            log.debug_traceback()
            return None

    def _gettag_id(self, tag_name):
        return self.con.execute("SELECT id FROM tags WHERE name = ?", [tag_name]).fetchone()[0]

    def _getsongs(self, sort=None, filters=None):
        """ returns songs filtered according to filters"""
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        orderstring = sort and sort.SQL_string() or ""
        args = filters and filters.SQL_args() or []
        select = """SELECT DISTINCT songs.id              AS song_id, 
                                    songs.album_id        AS album_id, 
                                    songs.artist_id       AS artist_id,
                                    songs.album_artist_id AS album_artist_id
                    FROM songs
                    LEFT JOIN artists   ON (songs.artist_id = artists.id)
                    LEFT JOIN albums    ON (songs.album_id = albums.id) 
                    %s
                    %s
                    %s
                    """ % (joinstring, wherestring, orderstring)
        # log.debug(select)
        return  [item.song(self.id, row["song_id"], row["album_id"], row["artist_id"], row["album_artist_id"])
                 for row in self.con.execute(select, args)]

    def _getartists(self, filters=None):
        """return artists filtered according to filters"""
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        args = filters and filters.SQL_args() or []
        select = """SELECT DISTINCT artists.id AS artist_id, artists.name AS artist_name
                    FROM artists 
                    JOIN songs         ON (songs.artist_id = artists.id)
                    LEFT JOIN albums   ON (album_id = albums.id)
                    %s
                    %s
                    ORDER BY artists.name COLLATE NOCASE""" % (joinstring, wherestring)
        # log.debug(select)
        return [item.artist(self.id, row["artist_id"], row["artist_name"], filters)
                for row in self.con.execute(select, args)]

    def _getalbums(self, filters=None):
        """return albums filtered according to filters"""
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        args = filters and filters.SQL_args() or []
        # Hackish, but effective to allow collections show up in artists view
        if filters.contains(item.artistfilter):
            artist_id_column = "artist_id"
        else:
            artist_id_column = "album_artist_id"
        select ="""SELECT DISTINCT albums.id AS album_id, artists.name AS artist_name, albums.name AS album_name
                   FROM albums 
                   JOIN artists  ON (songs.%s = artists.id)
                   JOIN songs    ON (songs.album_id = albums.id)
                   %s
                   %s
                   ORDER BY albums.name COLLATE NOCASE""" % (artist_id_column, joinstring, wherestring)

        # log.debug(select)
        return [item.album(self.id, row["album_id"], row["artist_name"], row["album_name"], filters)
                for row in self.con.execute(select, args)]

    def _gettags(self, filters=None):
        """return tags filtered according to filters"""
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        args = filters and filters.SQL_args() or []
        select ="""SELECT DISTINCT tags.id AS tag_id, tags.name AS tag_name
                   FROM tags
                   JOIN taggings ON (taggings.tag_id = tags.id)
                   JOIN songs ON (songs.id = taggings.song_id)
                   %s
                   %s
                   ORDER BY tags.name COLLATE NOCASE""" % (joinstring, wherestring)
        # log.debug(select)
        return [item.tag(self.id, row["tag_id"], row["tag_name"], filters)
                for row in self.con.execute(select, args)]

    def _getratings(self, filters):
        """return all stored ratings"""
        return []

    def _getlastplayedsongs(self, sort=None, filters=None):
        """return the last played songs"""
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        orderstring = sort and sort.SQL_string() or ""
        args = filters and filters.SQL_args() or []
        select = """SELECT DISTINCT songs.id              AS song_id,
                                    songs.album_id        AS album_id,
                                    songs.artist_id       AS artist_id,
                                    songs.album_artist_id AS album_artist_id,
                                    playstats.date_played AS date_played
                    FROM songs
                    LEFT JOIN artists   ON (songs.artist_id = artists.id)
                    LEFT JOIN albums    ON (songs.album_id = albums.id) 
                    JOIN      playstats ON (songs.id = playstats.song_id)
                    %s
                    %s
                    %s
                    """ % (joinstring, wherestring, orderstring)
        # log.debug(select)
        return  [item.song(self.id, row["song_id"], row["album_id"], row["artist_id"], 
                           row["album_artist_id"], row["date_played"])
                 for row in self.con.execute(select, args)]

    def _getplaylists(self, filters=None):
        joinstring = filters and filters.SQL_JOIN_string() or ""
        wherestring = filters and filters.SQL_WHERE_string() or ""
        args = filters and filters.SQL_args() or []
        select ="""SELECT DISTINCT playlists.id AS playlist_id, playlists.name AS playlist_name
                   FROM playlists
                   JOIN playlistcontents ON (playlistcontents.playlist_id = playlists.id)
                   JOIN songs ON (songs.id = playlistcontents.song_id)
                   %s
                   %s
                   ORDER BY playlists.name COLLATE NOCASE""" % (joinstring, wherestring)
        # JOIN taggings ON (taggings.tag_id = tags.id)
        # log.debug(select)
        return [item.playlist(self.id, row["playlist_id"], row["playlist_name"], filters)
                for row in self.con.execute(select, args)]

    def isbusy(self):
        """ check whether db is currently busy """
        return self.cur is not None or self.channel.queue.qsize()>0

    # event handlers

    def add_song(self, event):
        if event.songdbid == self.id:
            try:
                self._add_song(event.song)
            except KeyError:
                log.debug_traceback()
                pass

    def update_song(self, event):
        if event.songdbid == self.id:
            try:
                self._update_song(event.song)
            except:
                log.debug_traceback()
                pass

    def delete_song(self, event):
        if event.songdbid == self.id:
            try:
                self._delete_song(event.song)
            except:
                log.debug_traceback()
                pass

    def song_played(self, event):
        if event.songdbid == self.id:
            try:
                self._song_played(event.song, event.date_played)
            except KeyError:
                pass

    def song_skipped(self, event):
        if event.songdbid == self.id:
            try:
                self._song_skipped(event.song)
            except KeyError:
                pass

    def add_playlist(self, event):
        if event.songdbid == self.id:
            try: 
                self._add_playlist(event.name, event.songs)
            except:
                pass

    def delete_playlist(self, event):
        if event.songdbid == self.id:
            try:
                self._delete_playlist(event.name)
            except KeyError:
                pass

    def update_playlist(self, event):
        if event.songdbid == self.id:
            try:
                self._update_playlist(event.name, event.songs)
            except KeyError:
                pass

    def clearstats(self, event):
        if event.songdbid == self.id:
            self._clearstats()

    # request handlers

    def getdatabasestats(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return songdbstats(self.id, "local", self.basedir, None, self.dbfile, self.cachesize, 
                           self.getnumberofsongs(request), 
                           self.getnumberofalbums(request),
                           self.getnumberofartists(request),
                           self.getnumberoftags(request))

    def getnumberofsongs(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self.con.execute("SELECT count(*) FROM songs").fetchone()[0]

    def getnumberoftags(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self.con.execute("SELECT count(*) FROM tags").fetchone()[0]

    def getnumberofratings(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return 0

    def getnumberofalbums(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self.con.execute("SELECT count(*) FROM albums").fetchone()[0]

    def getnumberofartists(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self.con.execute("SELECT count(*) FROM artists").fetchone()[0]

    def getsong_metadata(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._getsong_metadata(song_id=request.song_id)
        except KeyError:
            return None

    def getsongs(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._getsongs(request.sort, request.filters)
        except (KeyError, AttributeError, TypeError):
            log.debug_traceback()
            return []

    def getartists(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._getartists(request.filters)
        except KeyError:
            log.debug_traceback()
            return []

    def getalbums(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._getalbums(request.filters)
        except KeyError:
            log.debug_traceback()
            return []

    def gettag_id(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._gettag_id(request.tag_name)
        except:
            log.debug_traceback()
            return None

    def gettags(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self._gettags(request.filters)

    def getplaylists(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self._getplaylists(request.filters)

    def getratings(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self._getratings(request.filters)

    def getlastplayedsongs(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        return self._getlastplayedsongs(request.sort, request.filters)


#
# thread for automatic registering and rescanning of songs in database
#

class songautoregisterer(service.service):

    def __init__(self, basedir, songdbid, dbbusymethod,
                 tracknrandtitlere, postprocessors):
        service.service.__init__(self, "songautoregisterer", daemonize=True)
        self.basedir = basedir
        self.songdbid = songdbid
        self.dbbusymethod = dbbusymethod
        self.tracknrandtitlere = tracknrandtitlere
        self.postprocessors = postprocessors
        self.done = False
        # support file extensions
        self.supportedextensions = metadata.getextensions()

        self.channel.subscribe(events.autoregistersongs, self.autoregistersongs)
        self.channel.subscribe(events.autoregisterer_rescansongs, self.autoregisterer_rescansongs)
        self.channel.supply(requests.autoregisterer_queryregistersong, self.autoregisterer_queryregistersong)

    def _notify(self, event):
        """ wait until db is not busy and send event """
        while self.dbbusymethod():
            time.sleep(0.1)
        hub.notify(event, -100)

    def _request(self, request):
        """ wait until db is not busy and send event """
        while self.dbbusymethod():
            time.sleep(0.1)
        return hub.request(request, -100)

    def _registerorupdatesong(self, path, force):
        """ register or update song in database and return it

        If force is set, the mtime of the song file is ignored.
        """
        if not path.startswith(self.basedir):
            log.error("Path of song not in basedir of database")
            return None

        # generate url corresponding to song
        if self.basedir.endswith("/"):
           relpath = path[len(self.basedir):]
        else:
           relpath = path[len(self.basedir)+1:]

        song_url = u"file://" + encoding.decode_path(relpath)
        urlfilter = item.filters((item.urlfilter(song_url),))
        songs = self._request(requests.getsongs(self.songdbid, filters=urlfilter))

        if songs:
            # there is exactly one resulting song
            song = songs[0]
            song.song_metadata = self._request(requests.getsong_metadata(self.songdbid, song.id))
            try:
                if force or song.song_metadata.date_updated < os.stat(path).st_mtime:
                    # the song has changed since the last update
                    newsong_metadata = metadata.metadata_from_file(relpath, self.basedir, self.tracknrandtitlere, self.postprocessors)
                    song.song_metadata.update(newsong_metadata)
                    self._notify(events.update_song(self.songdbid, song))
                    log.debug("registerer: song '%r' rescanned" % song_url)
                else:
                    log.debug("registerer: not scanning unchanged song '%r'" % song_url)
            except (IOError, OSError, RuntimeError):
                log.debug("registerer: song '%r' can no longer be read. deleting it from db" % song_url)
                self._notify(events.delete_song(self.songdbid, song))
        else:
            # song was not stored in database
            newsong_metadata = metadata.metadata_from_file(relpath, self.basedir, self.tracknrandtitlere, self.postprocessors)
            self._notify(events.add_song(self.songdbid, newsong_metadata))
            # fetch new song from database
            song = self._request(requests.getsongs(self.songdbid, filters=urlfilter))[0]
        return song

    def registerdirtree(self, dir, oldsongs, force):
        """ scan for songs in dir and its subdirectories, removing those scanned from the set oldsongs. 

        If force is set, the m_time of a song is ignored and the song is always scanned.
        """
        log.debug("registerer: entering %r"% dir)
        self.channel.process()
        if self.done: return
        songpaths = []

        # scan for paths of songs and recursively call registering of subdirectories
        for name in os.listdir(dir):
            path = os.path.join(dir, name)
            extension = os.path.splitext(path)[1].lower()
            if os.access(path, os.R_OK):
                if os.path.isdir(path):
                    try:
                        self.registerdirtree(path, oldsongs, force)
                    except (IOError, OSError), e:
                        log.warning("songautoregisterer: could not enter dir %r: %r" % (path, e))
                elif extension in self.supportedextensions:
                    songpaths.append(path)

        # now register songs...
        songs = []
        for path in songpaths:
            try:
                song = self._registerorupdatesong(path, force)
                # remove song from list of songs to be checked (if present)
                oldsongs.discard(song)
            except (IOError, OSError):
                # if the registering or update failed we do nothing and the song
                # will be deleted from the database later on
                pass
            except:
                # but in case of non-IO exceptions report them in debugging mode
                log.debug_traceback()
        log.debug("registerer: leaving %r"% dir)

    def run(self):
        # wait a little bit to not disturb the startup too much
        time.sleep(2)
        service.service.run(self)

    def rescansong(self, song, force):
        if song.songdbid != self.songdbid:
            log.debug("Trying to rescan song in wrong database")
            return
        if song.song_metadata is None:
            song.song_metadata = self._request(requests.getsong_metadata(self.songdbid, song.id))
            if song.song_metadata is None:
                log.debug("Song not found in database")
                return
        if not song.url.startswith("file://"):
            log.debug("Can only rescan local files")
            return
        relpath = encoding.encode_path(song.url[7:])
        path = os.path.join(self.basedir, relpath)
        try:
            if force or song_metadata.date_updated < os.stat(path).st_mtime:
                newsong_metadata = metadata.metadata_from_file(relpath, self.basedir, self.tracknrandtitlere, self.postprocessors)
                song.song_metadata.update(newsong_metadata)
                self._notify(events.update_song(self.songdbid, song))
        except (IOError, OSError):
            log.debug_traceback()
            # if anything goes wrong, we delete the song from the database
            self._notify(events.delete_song(self.songdbid, song))

    #
    # event handler
    #

    def autoregistersongs(self, event):
        if self.songdbid == event.songdbid:
            oldsongs = set(hub.request(requests.getsongs(self.songdbid)))
            log.info(_("database %r: scanning for songs in %r (currently %d songs registered)") % (self.songdbid, self.basedir, len(oldsongs)))

            # scan for all songs in the filesystem
            log.debug("database %r: searching for new songs" % self.songdbid)
            self.registerdirtree(self.basedir, oldsongs, event.force)

            # remove songs which have not yet been scanned and thus are not accesible anymore
            log.info(_("database %r: removing %d stale songs") % (self.songdbid, len(oldsongs)))
            for song in oldsongs:
                self._notify(events.delete_song(self.songdbid, song))

            nrsongs = hub.request(requests.getnumberofsongs(self.songdbid))
            log.info(_("database %r: rescan finished (%d songs registered)") % (self.songdbid, nrsongs))

    def autoregisterer_rescansongs(self, event):
        if self.songdbid == event.songdbid:
            log.info(_("database %r: rescanning %d songs%s") % (self.songdbid, len(event.songs), event.force and _(" (full)") or ""))
            for song in event.songs:
                self.rescansong(song, event.force)
            log.info(_("database %r: finished rescanning %d songs") % (self.songdbid, len(event.songs)))

    def autoregisterer_queryregistersong(self, request):
        if self.songdbid != request.songdbid:
            raise hub.DenyRequest
        try:
            return self._registerorupdatesong(request.path, force=False)
        except:
            return None



import sys
import locale

_fallbacklocalecharset = "iso-8859-1"

try:
    # works only in python > 2.3
    _localecharset = locale.getpreferredencoding()
except:
    try:
        _localecharset = locale.getdefaultlocale()[1]
    except:
        try:
            _localecharset = sys.getdefaultencoding()
        except:
            _localecharset = _fallbacklocalecharset
if _localecharset in [None, 'ascii', 'ANSI_X3.4-1968']:
    _localecharset = _fallbacklocalecharset

_fs_encoding = sys.getfilesystemencoding()
if _fs_encoding in [None, 'ascii', 'ANSI_X3.4-1968']:
    _fs_encoding = _fallbacklocalecharset

# exported functions

def encode(ustring):
    return ustring.encode(_localecharset, "replace")

def decode(string):
    return string.decode(_localecharset, "replace")

def decode_path(path):
    return path.decode(_fs_encoding)

def encode_path(path):
    return path.encode(_fs_encoding)



# -*- coding: UTF-8 -*-
##############################################################################
#
# Copyright (c) 2004 Nicolas Évrard All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import events
import config
import log
import plugin
import scrobbler

# the configuration section [plugin.audioscrobbler] contains two options:
# username and password

class audioscrobblerconfig(config.configsection):
    username = config.configstring("")
    password = config.configstring("")


# sending of audioscrobbler information may take time, so we make this a
# separate thread by deriving from plugin.threadedplugin

class audioscrobblerplugin(plugin.threadedplugin):

    def init(self):
        self.scrobbler = scrobbler.Scrobbler(self.config.username, self.config.password)
        self.scrobbler.handshake()
        self.lastSong = None
        self.channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
        log.info("started audiscrobbler plugin")

    def playbackinfochanged(self, event):
        if not event.playbackinfo.isplaying():
            return
        song = event.playbackinfo.song
        # do not submit songs which are shorter than 30 seconds or longer than 30 minutes
        if song.length < 30 or song.length > 30*60:
            return
        # submit when the song playback is 50% complete or 240 seconds have been passed,
        # whatever comes first
        mintime = (song.length <= 480 and int(song.length/2)) or 240
        if ( event.playbackinfo.time >= mintime and song != self.lastSong):
            log.debug("Audioscrobbler: submitting song '%s'" % song.id)
            try:
                self.scrobbler.submit(song)
                log.debug("Audioscrobbler: submission successful")
            except scrobbler.SubmissionError:
                log.error("Audioscrobbler: submission failed")
            except scrobbler.BadAuthError:
                log.error("Audioscrobbler: incorrect account information")
            self.lastSong = song



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2006, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import item
import services.playlist
import events
import messagewin
import encoding

class lyricswin(messagewin.messagewin):

    def __init__(self, screen, maxh, maxw, channel):
        messagewin.messagewin.__init__(self, screen, maxh, maxw, channel,
                                       config.colors.lyricswindow,
                                       _("Lyrics"), [],
                                       config.lyricswindow.autoclosetime)
        self.lyrics = _("No lyrics")
        channel.subscribe(events.selectionchanged, self.selectionchanged)

    def _outputlen(self, width):
        try:
            return len(self.lyrics[0][2].split("\n"))
        except IndexError:
            return 1

    def showitems(self):
        self.clear()
        try:
            for lno, line in enumerate(self.lyrics[0][2].split("\n")[self.first:self.first+self.ih]):
                line = encoding.encode(line).strip().center(self.iw)
                self.addnstr(self.iy+lno, self.ix, line, self.iw, self.colors.content)
        except IndexError:
            pass

    def selectionchanged(self, event):
        if isinstance(event.item, item.song):
            song = event.item
        elif isinstance(event.item, services.playlist.playlistitem):
            song = event.item.song
        else:
             self.settitle(_("Lyrics"))
             self.lyrics = _("No song selected")
             return
        self.settitle("%s - %s - %s" % (song.artist, song.album, song.title))
        if song.lyrics:
            self.lyrics = song.lyrics
        else:
            self.lyrics = _("No lyrics")



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import os.path, string, time
import config, metadata
import events, hub, requests
import encoding
import helper

#
# filters
#

class filter:
    def __init__(self, name, indexname, indexid):
        self.name = name
        self.indexname = indexname
        self.indexid = indexid

    def __repr__(self):
        # for dbrequest cache
        return "%r=%r" % (self.indexname, self.indexid)

    def SQL_JOIN_string(self):
        return ""

    def SQL_WHERE_string(self):
        return ""

    def SQL_args(self):
        return []


class hiddenfilter(filter):
    " a filter which does not show up in the UI "
    def __init__(self, indexname, indexid):
        filter.__init__(self, None, indexname, indexid)


class urlfilter(hiddenfilter):
    def __init__(self, url):
        self.url = url
        hiddenfilter.__init__(self, "url", url)

    def SQL_WHERE_string(self):
        return "songs.url = ?"

    def SQL_args(self):
        return [self.url]


class compilationfilter(hiddenfilter):
    def __init__(self, iscompilation):
        self.iscompilation = iscompilation
        hiddenfilter.__init__(self, "compilation", iscompilation)

    def SQL_WHERE_string(self):
        return "%s songs.compilation" % (not self.iscompilation and "NOT" or "")
        # return "(songs.compilation = %s)" % (self.iscompilation and "1" or "0")


class artistfilter(hiddenfilter):
    def __init__(self, artist_id):
        self.artist_id = artist_id
        hiddenfilter.__init__(self, "artist_id", artist_id)

    def SQL_WHERE_string(self):
        return "artists.id = ? OR songs.album_artist_id = ?"

    def SQL_args(self):
        return [self.artist_id, self.artist_id]


class noartistfilter(hiddenfilter):
    def __init__(self):
        hiddenfilter.__init__(self, "artist_id", None)

    def SQL_WHERE_string(self):
        return "songs.artist_id IS NULL"


class albumfilter(hiddenfilter):
    def __init__(self, album_id):
        self.album_id = album_id
        hiddenfilter.__init__(self, "album_id", album_id)

    def SQL_WHERE_string(self):
        return "albums.id = ?"

    def SQL_args(self):
        return [self.album_id]


class playlistfilter(hiddenfilter):
    def __init__(self, playlist_id):
        self.playlist_id = playlist_id
        hiddenfilter.__init__(self, "playlist_id", playlist_id)

    def SQL_JOIN_string(self):
        return "JOIN playlistcontents ON playlistcontents.song_id = songs.id"

    def SQL_WHERE_string(self):
        return "playlistcontents.playlist_id = ?"

    def SQL_args(self):
        return [self.playlist_id]


class playedsongsfilter(filter):
    def __init__(self):
        filter.__init__(self, _("Played songs"), "playedsongs", "true")

    def SQL_WHERE_string(self):
        return "songs.playcount > 0"


class searchfilter(filter):
    def __init__(self, searchstring):
        self.searchstring = searchstring
        filter.__init__(self, "Search: %s" % searchstring, None, searchstring)

    def SQL_WHERE_string(self):
        #return "(songs.title LIKE ?)"
        return "(songs.title LIKE ?) OR (albums.name LIKE ?) OR (artists.name LIKE ?)"

    def SQL_args(self):
        return ["%%%s%%" % self.searchstring] * 3


class tagfilter(filter):

    """ filters only items of given tag,
    
    if tag_id is not None, use this id to query the database, otherwise use tag_name """

    def __init__(self, tag_name, tag_id=None, inverted=False):
        self.tag_name = tag_name
        self.tag_id = tag_id
        self.inverted = inverted
        name = "%s%s=%s" % (_("Tag"), inverted and "!" or "", tag_name)
        filter.__init__(self, name, indexname="tag", indexid=tag_name)

    def __repr__(self):
        return "tag%r=%r" % (self.inverted and "!" or "", self.tag_name)

    def SQL_WHERE_string(self):
        if self.tag_id:
            return ( "songs.id %sIN (SELECT taggings.song_id FROM taggings WHERE taggings.tag_id = %d)" % 
                     (self.inverted and "NOT " or "", self.tag_id) )
        else:
            return ( "songs.id %sIN (SELECT taggings.song_id FROM taggings, tags WHERE taggings.tag_id = tags.id and tags.name=?)" % 
                     (self.inverted and "NOT " or "") )

    def SQL_args(self):
        if not self.tag_id:
            return [self.tag_name]
        else:
            return []


class podcastfilter(tagfilter):
    def __init__(self, inverted=False):
        tagfilter.__init__(self, "G:Podcast", inverted=inverted)
        # hide filter
        self.name = None


class deletedfilter(tagfilter):
    def __init__(self, inverted=False):
        tagfilter.__init__(self, "S:Deleted", inverted=inverted)
        # hide filter
        self.name = None


class ratingfilter(filter):

    """ filters only items of given rating """

    def __init__(self, rating):
        if rating is not None:
            name = "%s=%s" % (_("Rating"), "*" * rating)
        else:
            name = "%s=%s" % (_("Rating"), _("Not rated"))
        self.rating = rating
        filter.__init__(self, name, indexname="rating", indexid=rating)

    def SQL_WHERE_string(self):
        if self.rating:
            return "songs.rating = ?"
        else:
            return "songs.rating IS NULL"

    def SQL_args(self):
        if self.rating:
            return [self.rating]
        else:
            return []


class filters(tuple):

    def getname(self):
        s = ", ".join([filter.name for filter in self if filter.name])
        if s:
            return " <%s>" % s
        else:
            return ""

    def added(self, filter):
        return filters(self + (filter,))

    def removed(self, filterclass):
        return filters(tuple([f for f in self if not isinstance(f, filterclass)]))

    def contains(self, filterclass):
        for f in self:
            if isinstance(f, filterclass):
                return True
        return False

    def SQL_JOIN_string(self):
        return "\n".join([filter.SQL_JOIN_string() for filter in self])

    def SQL_WHERE_string(self):
        wheres = [filter.SQL_WHERE_string() for filter in self]
        wheres = ["(%s)" % s for s in wheres if s]
        filterstring = " AND ".join(wheres)
        if filterstring:
            filterstring = "WHERE (%s)" % filterstring
            return filterstring

    def SQL_args(self):
        result = []
        for filter in self:
            result.extend(filter.SQL_args())
        return result

# helper function for usage in getinfo methods, which merges information about
# filters in third and forth columns of lines
def _mergefilters(lines, filters):
    # filter out filters which are to be shown
    filters = [filter for filter in filters if filter.name]
    if filters:
        for nr, filter in enumerate(filters[:4]):
            if len(lines) > nr:
                lines[nr][2:3] = [_("Filter:"), filter.name]
            else:
                lines.append(["", "", _("Filter:"), filter.name])
    return lines


class item(object):
    """ base class for various items presentend in the database and
    playlist windows."""

    def __init__(self, songdbid, id):
        """ each item has to be bound to a specific database
        identified by songdbid """
        self.songdbid = songdbid
        self.id = id

    def getid(self):
        """ return unique id of item in context """
        raise NotImplementedError("has to be implemented by sub classes")

    def getname(self):
        """ short name used for item in lists """
        raise NotImplementedError("has to be implemented by sub classes")

    def getinfo(self):
        """ 4x4 array containing rows and columns used for display of item
        in iteminfowin"""
        return [["", "", "", ""]]

    def getinfolong(self):
        """ nx4 array containing rows and columns used for display of item
        in iteminfowin2"""
        return self.getinfo()

class diritem(item):

    """ item containing other items """

    def getname(self):
        return "[%s]/" % self.name

    def getid(self):
        return self.name

    def getcontents(self):
        """ return items contained in self """
        pass

    def getcontentsrecursive(self):
        """ return items contained in self including subdirs (in arbitrary order)"""
        result = []
        for aitem in self.getcontents():
            if isinstance(aitem, diritem):
                result.extend(aitem.getcontentsrecursive())
            else:
                result.append(aitem)

        return result

    def getcontentsrecursivesorted(self):
        """ return items contained in self including subdirs (sorted)"""
        result = []
        for aitem in self.getcontents():
            if isinstance(aitem, diritem):
                result.extend(aitem.getcontentsrecursivesorted())
            else:
                result.append(aitem)

        return result

    def getcontentsrecursiverandom(self):
        """ return random list of items contained in self including subdirs """
        # this should be implemented by subclasses
        return []

    def getheader(self, item):
        """ return header (used for title bar in filelistwin) of item in self.

        Note that item can be None!
        """
        if item and item.artist and item.album:
            s = "%s - %s" % (item.artist, item.album)
        else:
            s = self.name
        return s + self.filters.getname()

    def getinfo(self):
        return _mergefilters([[self.name, "", "", ""]], self.filters)


#
# specialized classes
#

def _formatnumbertotal(number, total):
    """ return string for number and total number """
    if number and total:
        return "%d/%d" % (number, total)
    elif number:
        return "%d" % number
    else:
        return "-"


class song(item):

    __slots__ = ["songdbid", "id", "album_id", "artist_id", "song", "playingtime"]

    def __init__(self, songdbid, id, album_id, artist_id, album_artist_id, date_played=None):
        """ create song with given id together with its database."""
        self.songdbid = songdbid
        self.id = id
        self.album_id = album_id
        self.artist_id = artist_id
        self.album_artist_id = album_artist_id
        self.date_played = date_played
        self.song_metadata = None

    def __repr__(self):
        return "song(%r) in %r database" % (self.id, self.songdbid)

    # the following two methods have to be defined because we use song as a
    # member of a set in the autoregisterer
    def __hash__(self):
        return hash("%r-%d" % (self.songdbid, self.id))

    def __eq__(self, other):
        return isinstance(other, song) and self.songdbid == other.songdbid and self.id == other.id

    def __getstate__(self):
        return (self.songdbid, self.id, self.album_id, self.artist_id, self.album_artist_id, self.date_played)

    def __setstate__(self, tuple):
        self.songdbid, self.id, self.album_id, self.artist_id, self.album_artist_id, self.date_played = tuple
        self.song_metadata = None

    def __getattr__(self, attr):
        # we refuse to fetch the song metadata if an "internal" method name is queried.
        # Thus, we do not interfere with pickling of song instances, etc.
        if attr.startswith("__"):
            raise AttributeError
        if not self.song_metadata:
            self.song_metadata = hub.request(requests.getsong_metadata(self.songdbid, self.id))
        # return metadata if we have been able to fetch it, otherwise return None
        if self.song_metadata:
            return getattr(self.song_metadata, attr)
        else:
            return None

    def _updatesong_metadata(self):
        """ notify database of song changes """
        hub.notify(events.update_song(self.songdbid, self))

    def getid(self):
        return self.id

    def getname(self):
        if self.title:
            return self.title
        else:
            return "DELETED"

    def getinfo(self):
        l = [["", "", "", ""]]*4
        # if we are unable to fetch the title, the song has been deleted in the meantime
        if self.title is None:
            return l
        l[0] = [_("Title:"), self.title]
        if self.tracknumber:
            l[0] += [_("Nr:"), _formatnumbertotal(self.tracknumber, self.trackcount)]
        else:
            l[0] += ["", ""]
        if self.album:
             l[1] = [_("Album:"),  self.album]
        else:
             l[1] = [_("URL:"), self.url]

        if self.year:
            l[1] += [_("Year:"), str(self.year)]
        else:
            l[1] += ["", ""]

        if self.artist:
            l[2] = [_("Artist:"), self.artist]
        else:
            l[2] = ["", ""]
        if self.length:
            l[2] += [_("Time:"), helper.formattime(self.length)]
        else:
            l[2] += ["", ""]
        if self.tags:
            l[3] = [_("Tags:"), u" | ".join(self.tags)]

        if self.getplayingtime() is not None:
            seconds = int((time.time()-self.getplayingtime())/60)
            days, rest = divmod(seconds, 24*60)
            hours, minutes = divmod(rest, 60)
            if days>=10:
                played = "%dd" % days
            elif days>0:
                played = "%dd %dh" % (days, hours)
            elif hours>0:
                played = "%dh %dm" % (hours, minutes)
            else:
                played = "%dm" % minutes
            if self.rating:
                played = played + " (%s)" % ("*"*self.rating)
            l[3] += [_("Played:"),
                   _("#%d, %s ago") % (self.playcount, played)]
        else:
            if self.rating:
                l[3] += [_("Rating:"), "*" * self.rating]
            else:
                l[3] += ["", ""]
        return l

    def getinfolong(self):
        l = []
        # if we are unable to fetch the title, the song has been deleted in the meantime
        if self.title is None:
            return l
        l.append([_("Title:"), self.title, "", ""])
        l.append([_("Album:"), self.album or "-", "", ""])
        l.append([_("Artist:"), self.artist or "-", "", ""])

        if self.year:
            year = str(self.year)
        else:
            year = "-"
        l.append([_("Time:"), "%d:%02d" % divmod(self.length, 60), _("Year:"), year])
        l.append([_("Track No:"), _formatnumbertotal(self.tracknumber, self.trackcount), 
                  _("Disk No:"), _formatnumbertotal(self.disknumber, self.diskcount)])
        l.append([_("Tags:"), u" | ".join(self.tags), _("Rating:"), self.rating and ("*" * self.rating) or "-"])

        if self.size:
            if self.size > 1024*1024:
                sizestring = "%.1f MB" % (self.size / 1024.0 / 1024)
            elif self.size > 1024:
                sizestring = "%.1f kB" % (self.size / 1024.0)
            else:
                sizestring = "%d B" % self.size
        else:
            sizestring = ""
        typestring = self.type.upper()
        if self.bitrate is not None:
            typestring = "%s %dkbps" % (typestring, self.bitrate/1000)
            if self.is_vbr:
                typestring = typestring + "VBR"
            if self.samplerate:
                typestring = "%s (%.1f kHz)" % (typestring, self.samplerate/1000.)

        l.append([_("File type:"), typestring, _("Size:"), sizestring])
        replaygain = ""
        if self.replaygain_track_gain is not None and self.replaygain_track_peak is not None:
            replaygain = replaygain + "%s: %+f dB (peak: %f) " % (_("track"),
                                                                  self.replaygain_track_gain,
                                                                  self.replaygain_track_peak)
        if self.replaygain_album_gain is not None and self.replaygain_album_peak is not None:
            replaygain = replaygain + "%s: %+f dB (peak: %f)" % (_("album"),
                                                                 self.replaygain_album_gain,
                                                                 self.replaygain_album_peak)
        l.append([_("Replaygain:"), replaygain or "-", _("Beats per minute:"), self.bpm and str(self.bpm) or "-"])
        l.append([_("Times played:"), str(self.playcount),_("Times skipped:"), str(self.skipcount)])
        l.append([_("Comment:"), self.comments and self.comments[0][2] or "-", 
                  _("Lyrics:"), self.lyrics and _("%d lines") % len(self.lyrics[0][2].split("\n")) or "-"])
        l.append([_("URL:"), self.url, "", ""])

        for played in self.dates_played[-1:-6:-1]:
            last = int((time.time()-played)/60)
            days, rest = divmod(last, 24*60)
            hours, minutes = divmod(rest, 60)
            if days>0:
                lastplayed = "%dd %dh %dm" % (days, hours, minutes)
            elif hours>0:
                lastplayed = "%dh %dm" % (hours, minutes)
            else:
                lastplayed = "%dm" % minutes

            l.append([_("Played:"), "%s (%s)" % (time.ctime(played), _("%s ago") % lastplayed), "", ""])

        return l

    def format(self, formatstring, adddict={}, safe=False):
        """format song info using formatstring. Further song information
        in adddict is added. If safe is True, all values are cleaned
        of characters which are neither letters, digits, a blank or a colon.
        """

        if self.title is None:
            return "DELETED"
        d = {}
        d.update(self.song_metadata.__dict__)
        d.update(adddict)
        d["minutes"], d["seconds"] = divmod(d["length"], 60)
        d["length"] = "%d:%02d" % (d["minutes"], d["seconds"])

        if safe:
            allowedchars = encoding.decode(string.letters + string.digits + " :")
            for key, value in d.items():
                try:
                    l = []
                    for c in value:
                        if c in allowedchars:
                            l.append(c)
                    d[key] = "".join(l)
                except TypeError:
                    pass

        return unicode(formatstring) % d

    def rate(self, rating):
        # just to fetch song metadata
        oldrating = self.rating
        # if this was sucessful we can rate the song
        if self.song_metadata:
            if rating:
                self.song_metadata.rating = rating
            else:
                self.song_metadata.rating = None
            self._updatesong_metadata()

    def addtag(self, tag):
        tags = self.tags
        if tags is not None and tag not in tags:
            tags.append(tag)
            self.tags = tags
            self._updatesong_metadata()

    def removetag(self, tag):
        tags = self.tags
        if tags is not None and tag in tags:
            tags.remove(tag)
            self.tags = tags
            self._updatesong_metadata()

    def toggledelete(self):
        if self.tags is None:
            return
        if "S:Deleted" in self.tags:
            self.tags.remove("S:Deleted")
        else:
            self.tags.append("S:Deleted")
        self._updatesong_metadata()

    def getplayingtime(self):
        """ return time at which this particular song instance has been played or the
        last playing time, if no such time has been specified at instance creation time """
        return self.date_played or self.date_lastplayed


class artist(diritem):

    """ artist bound to specific songdb """

    def __init__(self, songdbid, id, name, filters):
        self.songdbid = songdbid
        self.id = id
        self.name = name

        self.filters = filters.removed(compilationfilter).added(artistfilter(id))

    def __repr__(self):
        return "artist(%r) in %r (filtered: %r)" % (self.name, self.songdbid, self.filters)

    def getname(self):
        return "%s/" % self.name

    def getcontents(self):
        albums = hub.request(requests.getalbums(self.songdbid, filters=self.filters))
        return albums + [songs(self.songdbid, self.name, self.filters)]

    def getcontentsrecursive(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters))

    def getcontentsrecursivesorted(self):
        albums = hub.request(requests.getalbums(self.songdbid, filters=self.filters))
        result = []
        for aalbum in albums:
            result.extend(aalbum.getcontentsrecursivesorted())
        return result

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters, random=True))

    def getheader(self, item):
        return self.name + self.filters.getname()

    def getinfo(self):
        if self.name == metadata.VARIOUS:
            # this should not happen, actually
            artistname = _("Various")
        else:
            artistname = self.name
        return _mergefilters([[_("Artist:"), artistname, "", ""]], self.filters)


class album(diritem):

    """ album bound to specific songdb """

    def __init__(self, songdbid, id, artist, name, filters):
        self.songdbid = songdbid
        self.id = id
        self.artist = artist
        self.name = name
        self.filters = filters.added(albumfilter(id))

    def __repr__(self):
        return "album(%r) in %r" % (self.id, self.songdbid)

    class _orderclass:
        def cmpitem(self, x, y):
            return ( x.disknumber and y.disknumber and cmp(x.disknumber, y.disknumber) or
                     x.tracknumber and y.tracknumber and cmp(x.tracknumber, y.tracknumber) or
                     cmp(x.title, y.title) )
        def SQL_string(self):
            return "ORDER BY songs.disknumber, songs.tracknumber, songs.title"
    order = _orderclass()

    def getid(self):
        return self.id

    def getname(self):
        return "%s/" % self.name

    def getcontents(self):
        songs = hub.request(requests.getsongs(self.songdbid, sort=self.order, filters=self.filters))
        return songs

    def getcontentsrecursive(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters))

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters, random=True))

    def getinfo(self):
        if self.artist == metadata.VARIOUS:
            artistname = _("Various")
        else:
            artistname = self.artist
        albumname =  self.name 
        l = [[_("Artist:"), artistname, "", ""],
             [_("Album:"), albumname, "", ""]]
        return _mergefilters(l, self.filters)


class playlist(diritem):

    """ songs in a playlist in the corresponding database """

    def __init__(self, songdbid, id, name, nfilters):
        self.songdbid = songdbid
        self.id = id
        self.name = name
        if nfilters is not None:
            self.filters = nfilters.added(playlistfilter(id))
        else:
            self.filters = filters((playlistfilter(id),))

    def getname(self):
        return "%s/" % self.name

    class _orderclass:
        def SQL_string(self):
            return "ORDER BY playlistcontents.position"
    order = _orderclass()

    def getcontents(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters, sort=self.order))

    getcontentsrecursive = getcontents

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters, random=True))

    def getheader(self, item):
        if item and item.artist and item.album:
            return item.artist + " - " + item.album
        else:
            return self.name

    def getinfo(self):
        return [["%s:" % _("Playlist"), self.name, "", ""]]


class totaldiritem(diritem):

    """ diritem which contains the total database(s) as its contents """

    def getcontentsrecursive(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters))

    getcontentsrecursivesorted = getcontentsrecursive

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, filters=self.filters, random=True))


class songs(totaldiritem):

    """ all songs in the corresponding database """

    def __init__(self, songdbid, artist=None, filters=None):
        self.songdbid = songdbid
        self.id = "songs"
        self.name = _("Songs")
        self.artist = artist
        self.filters = filters
        self.nrsongs = None

    def getname(self):
        if self.nrsongs is None:
            self.nrsongs = hub.request(requests.getnumberofsongs(self.songdbid, filters=self.filters))
        return "[%s (%d)]/" % (self.name, self.nrsongs)

    class _orderclass:
        def cmpitem(self, x, y):
            return ( cmp(x.title, y.title) or
                     cmp(x.album, y.album) or
                     cmp(x.path, y.path)
                     )
        def SQL_string(self):
            return "ORDER BY songs.title, albums.name, songs.url"
    order = _orderclass()

    def getcontents(self):
        songs = hub.request(requests.getsongs(self.songdbid, filters=self.filters, sort=self.order))
        self.nrsongs = len(songs)
        return songs

    def getinfo(self):
        if self.artist is not None:
            l = [[_("Artist:"), self.artist, "", ""],
                    [self.name, "", "", ""]]
        else:
            l = [[self.name, "", "", ""]]
        return _mergefilters(l, self.filters)


class noartist(songs):

    """ list of songs without artist information """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "noartist"
        self.name = _("No artist")
        self.filters = filters.added(noartistfilter())
        self.nrsongs = None

    def getinfo(self):
        return _mergefilters([[self.name, "", "", ""]], self.filters)


class randomsongs(totaldiritem):

    """ random list of songs out of  the corresponding database """

    def __init__(self, songdbid, maxnr, filters):
        self.songdbid = songdbid
        self.id = "randomsongs"
        self.name = _("Random song list")
        self.maxnr = maxnr
        self.filters = filters

    def getcontents(self):
        songs = []
        while len(songs)<self.maxnr:
            newsongs = hub.request(requests.getsongs(self.songdbid, filters=self.filters, random=True))
            if len(newsongs) > 0:
                songs.extend(newsongs)
            else:
                break
        return songs[:self.maxnr]


class lastplayedsongs(diritem):

    """ songs last played out of the corresponding databases """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "lastplayedsongs"
        self.filters = filters.added(playedsongsfilter())
        self.name = _("Last played songs")

    class _orderclass:
        def cmpitem(self, x, y):
            return cmp(y.getplayingtime(), x.getplayingtime())
        def SQL_string(self):
            return "ORDER BY playstats.date_played DESC LIMIT 100"
    order = _orderclass()

    def getcontents(self):
        return hub.request(requests.getlastplayedsongs(self.songdbid, sort=self.order, filters=self.filters))

    getcontentsrecursive = getcontentsrecursivesorted = getcontents

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getlastplayedsongs(self.songdbid, filters=self.filters, random=True))

    def getinfo(self):
        return _mergefilters([[self.name, "", "", ""]], self.filters[:-1])


class topplayedsongs(diritem):

    """ songs most often played of the corresponding databases """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "topplayedsongs"
        self.filters = filters.added(playedsongsfilter())
        self.name = _("Top played songs")

    class _orderclass:
        def cmpitem(self, x, y):
            return cmp(y.playcount, x.playcount) or cmp(y.date_lastplayed, x.date_lastplayed)
        def SQL_string(self):
            return "ORDER BY songs.playcount DESC, songs.date_lastplayed DESC LIMIT 100"
    order = _orderclass()

    def getcontents(self):
        songs = hub.request(requests.getsongs(self.songdbid, sort=self.order, filters=self.filters))
        return songs

    getcontentsrecursive = getcontentsrecursivesorted = getcontents

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, sort=self.order, filters=self.filters, random=True))

    def getinfo(self):
        return _mergefilters([[self.name, "", "", ""]], self.filters[:-1])


class lastaddedsongs(diritem):

    """ songs last added to the corresponding database """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "lastaddedsongs"
        self.filters = filters
        self.name = _("Last added songs")

    class _orderclass:
        def cmpitem(self, x, y):
            return cmp(y.date_added, x.date_added)
        def SQL_string(self):
            return "ORDER BY songs.date_added DESC LIMIT 100"
    order = _orderclass()

    def getcontents(self):
        return hub.request(requests.getsongs(self.songdbid, sort=self.order, filters=self.filters))

    getcontentsrecursive = getcontentsrecursivesorted = getcontents

    def getcontentsrecursiverandom(self):
        return hub.request(requests.getsongs(self.songdbid, sort=self.order, filters=self.filters, random=True))


class albums(totaldiritem):

    """ all albums in the corresponding database """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "albums"
        self.filters = filters
        self.name = _("Albums")
        self.nralbums = None

    def getname(self):
        if self.nralbums is None:
            self.nralbums = hub.request(requests.getnumberofalbums(self.songdbid, filters=self.filters))
        return "[%s (%d)]/" % (self.name, self.nralbums)

    def getcontents(self):
        albums = hub.request(requests.getalbums(self.songdbid, filters=self.filters))
        self.nralbums = len(albums)
        return albums

    def getheader(self, item):
        if self.nralbums is None:
            self.nralbums = len(self.getcontents())
        return "%s (%d)" % (self.name, self.nralbums) + self.filters.getname()


class compilations(albums):
    def __init__(self, songdbid, filters):
        filters = filters.added(compilationfilter(True))
        albums.__init__(self, songdbid, filters)
        self.id = "compilations"
        self.name = _("Compilations")


class podcasts(albums):
    def __init__(self, songdbid, filters):
        filters = filters.removed(podcastfilter)
        filters = filters.added(podcastfilter())
        albums.__init__(self, songdbid, filters)
        self.id = "podcasts"
        self.name = _("Podcasts")


class deleted(albums):
    def __init__(self, songdbid, filters):
        filters = filters.removed(deletedfilter)
        filters = filters.added(deletedfilter())
        import log
        log.debug(str(filters))
        albums.__init__(self, songdbid, filters)
        self.id = "deleted"
        self.name = _("Deleted songs")


class tags(totaldiritem):

    """ all tags in the corresponding database """

    def __init__(self, songdbid, songdbids, filters):
        self.songdbid = songdbid
        self.id = "tags"
        self.songdbids = songdbids
        self.filters = filters
        self.name = _("Tags")
        self.nrtags = None
        self.exclude_tag_ids = []
        for filter in self.filters:
            if isinstance(filter, tagfilter):
                if not filter.inverted:
                    self.exclude_tag_ids.append(filter.tag_id)

    def getname(self):
        if self.nrtags is None:
            self.nrtags = len(self.getcontents())
        return "[%s (%d)]/" % (self.name, self.nrtags)

    def getcontents(self):
        tags = hub.request(requests.gettags(self.songdbid, filters=self.filters))
        tags = [tag for tag in tags if tag.id not in self.exclude_tag_ids]
        self.nrtags = len(tags)
        return tags

    def getheader(self, item):
        if self.nrtags is None:
            self.nrtags = len(self.getcontents())
        return "%s (%d)" % (self.name, self.nrtags) + self.filters.getname()


class ratings(totaldiritem):

    """ all ratings in the corresponding database """

    def __init__(self, songdbid, songdbids, filters):
        self.songdbid = songdbid
        self.id = "ratings"
        self.songdbids = songdbids
        self.filters = filters
        self.name = _("Ratings")
        self.nrratings = 6

    def getname(self):
        if self.nrratings is None:
            self.nrratings = len(self.getcontents())
        return "[%s (%d)]/" % (self.name, self.nrratings)

    def getcontents(self):
        ratings = [rating(self.songdbid, r, self.filters) for r in range(1, 6)]
        ratings.append(rating(self.songdbid, None, self.filters))
        self.nrratings = len(ratings)
        return ratings

    def getheader(self, item):
        if self.nrratings is None:
            nrratings = len(self.getcontents())
        return "%s (%d)" % (self.name, self.nrratings) + self.filters.getname()


class playlists(diritem):

    """ all playlists in the corresponding database """

    def __init__(self, songdbid, filters):
        self.songdbid = songdbid
        self.id = "playlists"
        self.name = _("Playlists")
        self.filters = filters
        self.nrplaylists = None

    def getname(self):
        if self.nrplaylists is None:
            self.nrplaylists = len(self.getcontents())
        return "[%s (%d)]/" % (_("Playlists"), self.nrplaylists)

    def getcontents(self):
        playlists = hub.request(requests.getplaylists(self.songdbid, filters=self.filters))
        self.nrplaylists = len(playlists)
        return playlists

    def getheader(self, item):
        if self.nrplaylists is None:
            self.nrplaylists = len(self.getcontents())
        return "%s (%d)" % (_("Playlists"), self.nrplaylists)


class filesystemdir(diritem):

    """ diritem corresponding to directory in filesystem """

    def __init__(self, songdbid, basedir, dir):
        self.songdbid = songdbid
        self.id = "filesystemdir"
        self.basedir = basedir
        self.dir = dir

        if self.dir==self.basedir:
            self.name =  _("Filesystem")
        else:
            self.name = encoding.decode_path(self.dir[len(self.basedir):].split("/")[-1])

    def getname(self):
        if self.isbasedir():
            return "[%s]/" % _("Filesystem")
        else:
            return "%s/" % self.name

    def getcontents(self):
        items = []
        try:
            for name in os.listdir(self.dir):
                try:
                    path = os.path.join(self.dir, name)
                    extension = os.path.splitext(path)[1]
                    if os.path.isdir(path) and os.access(path, os.R_OK|os.X_OK):
                        newitem = filesystemdir(self.songdbid, self.basedir, path)
                        items.append(newitem)
                    elif extension in metadata.getextensions() and os.access(path, os.R_OK):
                        song = hub.request(requests.autoregisterer_queryregistersong(self.songdbid, path))
                        if song:
                            items.append(song)
                except (IOError, OSError): pass
        except OSError:
            return None
        items.sort(cmp=lambda x, y: cmp(x.getname(), y.getname()))
        return items

    def getcontentsrecursiverandom(self):
        return []
        # songs = self.getcontentsrecursive()
        # return _genrandomchoice(songs)

    def getheader(self, item):
        if self.isbasedir():
            return _("Filesystem")
        else:
            return self.name

    def getinfo(self):
        return [["%s:" % _("Filesystem"), encoding.decode_path(self.dir), "", ""]]

    def isbasedir(self):
        """ return whether the filesystemdir is the basedir of a song database """
        return self.dir == self.basedir

_dbstats = None

class basedir(totaldiritem):

    """ base dir of database view"""

    def __init__(self, songdbids, afilters=None, rootdir=False):
        # XXX: as a really dirty hack, we cache the result of getdatabasestats for
        # all databases because we cannot call this request safely later on
        # (we might be handling another request which calls the basedir constructor)
        global _dbstats
        if _dbstats is None:
            _dbstats = {}
            for songdbid in songdbids:
                _dbstats[songdbid] = hub.request(requests.getdatabasestats(songdbid))
        self.name =  _("Song Database")
        self.songdbids = songdbids
        if len(songdbids) == 1:
            self.songdbid = songdbids[0]
            self.type = _dbstats[self.songdbid].type
            self.basedir = _dbstats[self.songdbid].basedir
        else:
            self.songdbid = None
            self.type = "virtual"
            self.basedir = None
        self.id = "basedir"
        if afilters is None:
            # add default filters
            self.filters = filters((podcastfilter(inverted=True),
                                    deletedfilter(inverted=True)))
        else:
            self.filters = afilters
        self.rootdir = rootdir
        self.maxnr = 100
        self.nrartists = None
        self.nrsongs = None
        self._initvirtdirs()

    def _initvirtdirs(self):
        self.virtdirs = []
        self.virtdirs.append(noartist(self.songdbid, filters=self.filters))
        self.virtdirs.append(compilations(self.songdbid, filters=self.filters))
        if self.type == "local" and self.rootdir:
            self.virtdirs.append(filesystemdir(self.songdbid, self.basedir, self.basedir))
        self.virtdirs.append(songs(self.songdbid, filters=self.filters))
        self.virtdirs.append(albums(self.songdbid, filters=self.filters))
        self.virtdirs.append(podcasts(self.songdbid, filters=self.filters))
        self.virtdirs.append(deleted(self.songdbid, filters=self.filters))

        if not self.filters.contains(searchfilter):
             self.virtdirs.append(tags(self.songdbid, self.songdbids, filters=self.filters))
        if not self.filters.contains(ratingfilter):
            self.virtdirs.append(ratings(self.songdbid, self.songdbids, filters=self.filters))
        if not self.filters.contains(playedsongsfilter):
            self.virtdirs.append(topplayedsongs(self.songdbid, filters=self.filters))
            self.virtdirs.append(lastplayedsongs(self.songdbid, filters=self.filters))
            self.virtdirs.append(playedsongs(self.songdbid, nfilters=self.filters))
        self.virtdirs.append(lastaddedsongs(self.songdbid, filters=self.filters))
        self.virtdirs.append(randomsongs(self.songdbid, self.maxnr, filters=self.filters))
        if not self.filters.contains(searchfilter):
            self.virtdirs.append(playlists(self.songdbid, filters=self.filters))
        if len(self.songdbids) > 1:
            self.virtdirs.extend([basedir([songdbid], self.filters) for songdbid in self.songdbids])

    def getname(self):
        if self.nrsongs is None:
            self.nrsongs = hub.request(requests.getnumberofsongs(self.songdbid, filters=self.filters))
        if self.basedir:
            return  _("[Database: %s (%d)]") % (self.basedir, self.nrsongs)
        else:
            return _("%d databases (%d)") % (len(self.songdbids), self.nrsongs)

    def getcontents(self):
        # do not show artists which only appear in compilations
        filters = self.filters.added(compilationfilter(False))
        aartists = hub.request(requests.getartists(self.songdbid, filters=filters))
        self.nrartists = len(aartists)
        # reset cached value
        self.nrsongs = None
        if config.filelistwindow.virtualdirectoriesattop:
            return self.virtdirs + aartists
        else:
            return aartists + self.virtdirs

    def getcontentsrecursivesorted(self):
        # we cannot rely on the default implementation since we don't want
        # to have the albums and songs included trice
        artists = hub.request(requests.getartists(self.songdbid, filters=self.filters))
        result = []
        for aartist in artists:
            result.extend(aartist.getcontentsrecursivesorted())
        return result

    def getheader(self, item):
        if self.nrartists is not None:
            nrartistsstring = _("%d artists") % self.nrartists
        else:
            nrartistsstring = _("? artists") 
        if self.basedir:
            maxlen = 15
            dirname = self.basedir
            if len(dirname)>maxlen:
                dirname = "..."+dirname[-maxlen+3:]
            else:
                dirname = self.basedir
            s = _("Database (%s, %s)") % (dirname, nrartistsstring)
        else:
            s = _("%d databases (%s)") % (len(self.songdbids), nrartistsstring)
        return s + self.filters.getname()

    def getinfo(self):
         if self.basedir:
             description = _("[Database: %s (%%d)]") % (self.basedir)
         else:
             description = _("%d databases (%%d)") % (len(self.songdbids))
         return _mergefilters([[self.name, description, "", ""]], self.filters)

class index(basedir):

    def __init__(self, songdbids, name, description, filters):
        basedir.__init__(self, songdbids, filters)
        self.name = name
        self.description = description
        self.type = "index"

    def getname(self):
        # XXX make this configurable (note that showing the numbers by default is rather costly)
        if 1:
            return "%s/" % self.description
        else:
            if self.nrsongs is None:
                self.nrsongs = hub.request(requests.getnumberofsongs(self.songdbid, filters=self.filters))
            return "%s (%d)/" % (self.description, self.nrsongs)

    def getinfo(self):
        return _mergefilters([[self.name, self.description, "", ""]], self.filters[:-1])


class tag(index):
    def __init__(self, songdbid, id, name, nfilters):
        if nfilters is not None:
            nfilters = nfilters.added(tagfilter(name, tag_id=id))
        else:
            nfilters = filters((tagfilter(name, tag_id=id),))
        index.__init__(self, [songdbid], _("Tag:"), name, nfilters)
        self.id = id


class rating(index):
    def __init__(self, songdbid, r, nfilters):
        if nfilters is not None:
            nfilters = nfilters.added(ratingfilter(r))
        else:
            nfilters = filters((ratingfilter(r),))
        if r is None:
            description = _("Not rated")
        else:
            description = "*" * r
        index.__init__(self, [songdbid], _("Rating:"), description, nfilters)
        self.id = r


class playedsongs(index):

    """ songs played at least once """

    def __init__(self, songdbid, nfilters):
        if nfilters is not None:
            nfilters = nfilters.added(playedsongsfilter())
        else:
            nfilters = filters((playedsongsfilter(),))
        index.__init__(self, [songdbid], _("Played songs:"), "[%s]" % _("Played songs"), nfilters)
        self.name = _("Played songs")
        self.id = "playedsongs"

    def getinfo(self):
        return _mergefilters([[self.name, "", "", ""]], self.filters[:-1])


class focus_on(index):

    """ songs filtered by search string """

    def __init__(self, songdbid, searchstring, nfilters):
        if nfilters is not None:
            nfilters = nfilters.added(searchfilter(searchstring))
        else:
            nfilters = filters((searchfilter(searchstring),))
        index.__init__(self, [songdbid], _("Search:"), searchstring, nfilters)
        self.id = "search: %s" % searchstring




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses

##############################################################################
# descriptions for functions and keys
#
# descriptions: dictionary with entries
#  "general":  global function descriptions
#  "filelist": filelist function descriptions
#  "playlist": playlistlist function descriptions
# which map from
#   name of function to 
#   2-tuple of very short (for statusbar) and short description of
#   corresponding function
#
# keyname:  dictionary which maps from  keycodes to the names of the keys
##############################################################################

descriptions = {
    "general": {
          "refresh":           (_("refresh"), _("refresh display")),
          "exit":              (_("exit"), _("exit PyTone (press twice)")), 
          "playerstart":       (_("play"), _("start main player")),
          "playerpause":       (_("pause"), _("pause main player")),
          "playernextsong":    (_("next song"), _("advance to next song")),
          "playerprevioussong":(_("previous song"), _("go back to previous song")),
          "playerrewind":      (_("rewind"), _("rewind main player")),
          "playerforward":     (_("forward"), _("forward main player")),
          "playerstop":        (_("stop"), _("stop main player")),
          "playlistdeleteplayedsongs":
                               (_("delete played"), _("delete played songs from playlist")),
          "playlistreplay":    (_("replay songs"), _("mark all songs in playlist as unplayed")),
          "playlisttoggleautoplaymode":
                               (_("toggle playlist mode"), _("toggle the playlist mode")),
          "playlistclear":     (_("clear"), _("clear playlist")),
          "playlistsave":      (_("save"), _("save playlist")),
          "showhelp":          (_("help"), _("show help")),
          "showlog":           (_("log"), _("show log messages")),
          "showstats":         (_("statistics"), _("show statistical information about database(s)")),
          "showiteminfolong":  (_("item info"), _("show information about selected item")),
          "showlyrics":        (_("lyrics"), _("show lyrics of selected song")),
          "toggleiteminfowindow":  (_("toggle item info"), _("toggle information shown in item info window")),
          "togglelayout":      (_("toggle layout"), _("toggle layout")),
          "volumeup":          (_("volume up"), _("increase output volume")),
          "volumedown":        (_("volume down"), _("decrease output volume")),
          "playerplayfaster":  (_("play faster"), _("increase the play speed")),
          "playerplayslower":  (_("play slower"), _("decrease the play speed")),
          "playerspeedreset":  (_("default play speed"), _("reset the play speed to normal")),
          "playerratecurrentsong1":  (_("rate current song 1"), _("rate currently playing song with 1 star")),
          "playerratecurrentsong2":  (_("rate current song 2"), _("rate currently playing song with 2 stars")),
          "playerratecurrentsong3":  (_("rate current song 3"), _("rate currently playing song with 3 stars")),
          "playerratecurrentsong4":  (_("rate current song 4"), _("rate currently playing song with 4 stars")),
          "playerratecurrentsong5":  (_("rate current song 5"), _("rate currently playing song with 5 stars")),
    },
    "filelistwindow": {
          "selectnext":        (_("down"), _("move to the next entry")),
          "selectprev":        (_("up"), _("move to the previous entry")),
          "selectnextpage":    (_("page down"), _("move to the next page")),
          "selectprevpage":    (_("page up"), _("move to previous page")),
          "selectfirst":       (_("first"), _("move to the first entry")),
          "selectlast":        (_("last"), _("move to the last entry")),
          "dirdown":           (_("enter dir"), _("enter selected directory")),
          "dirup":             (_("exit dir"), _("go directory up")),
          "addsongtoplaylist": (_("add song"), _("add song to playlist")),
          "adddirtoplaylist":  (_("add dir"), _("add directory recursively to playlist")),
          "playselectedsong":  (_("immediate play"), _("play selected song immediately")),
          "activateplaylist":  (_("switch to playlist"), _("switch to playlist window")),
#          "generaterandomlist":(_("random suggestion"), _("generate random song list")),
          "insertrandomlist":  (_("random add dir"), _("add random contents of dir to playlist")),
          "search":            (_("search"), _("search entry")),
          "repeatsearch":      (_("repeat search"), _("repeat last search")),
          "focus":             (_("focus"), _("focus by search string")),
          "rescan":            (_("rescan"), _("rescan/update id3 info for selection")),
          "toggledelete":      (_("(un)delete"), _("delete/undelete selection ")),
          },
    "playlistwindow": {
          "selectnext":        (_("down"), _("move to the next entry")),
          "selectprev":        (_("up"), _("move to the previous entry")),
          "selectnextpage":    (_("page down"), _("move to the next page")),
          "selectprevpage":    (_("page up"), _("move to previous page")),
          "selectfirst":       (_("first"), _("move to the first entry")),
          "selectlast":        (_("last"), _("move to the last entry")),
          "moveitemup":        (_("move song up"), _("move song up")),  
          "moveitemdown":      (_("move song down"), _("move song down")),
          "shuffle":           (_("shuffle"), _("shuffle playlist")),
          "deleteitem":        (_("delete"), _("delete entry")),
          "playselectedsong":  (_("immediate play"), _("play selected song immediately")),
          "activatefilelist":  (_("switch to database"), _("switch to database window")),
          "rescan":            (_("rescan"), _("rescan/update id3 info for selection")),
          "filelistjumptoselectedsong":
                               (_("jump to selected"), _("jump to selected song in filelist window")),
          }
    }

# prefill keynames
keynames = {}
for c in range(32):
    keynames[c] = _("CTRL")+"-"+chr(c+64)
for c in range(33, 128):
    keynames[c] = chr(c)


# special keys (incomplete, but sufficient list)
keynames[ord("\t")]            = _("<TAB>")
keynames[ord("\n")]            = _("<Return>")
keynames[27]                   = _("<ESC>")
keynames[32]                   = _("<Space>")
keynames[curses.KEY_BACKSPACE] = _("<Backspace>")
keynames[curses.KEY_DC]        = _("<Del>")
keynames[curses.KEY_DOWN]      = _("<Down>")
keynames[curses.KEY_END]       = _("<End>")
keynames[curses.KEY_ENTER]     = _("<Enter>")
keynames[curses.KEY_HOME]      = _("<Home>")
keynames[curses.KEY_IC]        = _("<Insert>")
keynames[curses.KEY_LEFT]      = _("<Left>")
keynames[curses.KEY_SLEFT]     = _("<Shift-Left>")
keynames[curses.KEY_NPAGE]     = _("<PageDown>")
keynames[curses.KEY_PPAGE]     = _("<PageUp>")
keynames[curses.KEY_RIGHT]     = _("<Right>")
keynames[curses.KEY_SRIGHT]    = _("<Shift-Right>")
keynames[curses.KEY_UP]        = _("<Up>")

# function keys
for nr in range(1, 23):
    keynames[eval("curses.KEY_F%d" % nr)] = "<F%d>" %nr

# alt+key
for key in keynames.keys():
    keynames[key+1024] = "%s-%s"% (_("Alt"), keynames[key])

def getkeyname(key):
    try:
        return keynames[key]
    except KeyError:
        return "Key_%d" % key 



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import Queue
import sys
import threading
import time

import hub, events
import pcm
import decoder
from services.player import genericplayer
import services.playlist
import log

try:
    import bufferedao
    import thread
    bufferedao_present = True
except ImportError:
    bufferedao_present = False
    try:
        import ao
        ao_present = True
    except ImportError:
        ao_present = False
    try:
        import ossaudiodev
        ossaudiodev_present = False
    except ImportError:
        ossaudiodev_present = False


class aoaudiodev:
    def __init__(self, aodevice, rate, options):
        self.ao = ao.AudioDevice(aodevice, rate=rate, byte_format=4, options=options)

    def play(self, buff, bytes):
        self.ao.play(buff, bytes)

    def close(self):
        # to close the ao audio device, we have to delete it...
        del self.ao 

# support for new ossaudiodev module of Python 2.3

class ossaudiodev:
    def __init__(self, device, rate):
        self.ossdevice = ossaudiodev.open(device, "w")
        if sys.byteorder == 'little':
            self.ossdevice.setfmt(ossaudiodev.AFMT_S16_LE)
        else:
            self.ossdevice.setfmt(ossaudiodev.AFMT_S16_BE)
        self.ossdevice.channels(2)
        self.ossdevice.speed(rate)

    def play(self, buff, bytes):
        self.ossdevice.write(buff)

    def close(self):
        self.ossdevice.close()


class bufferedaudiodev(threading.Thread):
    def __init__(self, aodevice, aooptions, bufsize, rate, SIZE):
        self.aodevice = aodevice
        self.aooptions = aooptions
        self.rate = rate
        self.SIZE = SIZE

        # initially, we do not open the audio device
        self.audiodev = None

        # output queue
        queuesize = 1024*bufsize/self.SIZE + 1
        self.queue = Queue.Queue(queuesize)

        self.done = False
        # wait if player thread is paused
        self.ispaused = False
        self.restart = threading.Event()
        threading.Thread.__init__(self)
        self.setDaemon(True)

    def opendevice(self):
        errorlogged = False
        while self.audiodev is None:
            try:
                if self.aodevice=="oss":
                    if ossaudiodev_present:
                        self.audiodev = ossaudiodev(self.aooptions["dsp"], self.rate)
                        log.debug("ossaudiodev audio device opened")
                    else:
                        self.audiodev = aoaudiodev(self.aodevice, rate=self.rate, options=self.aooptions)
                        log.debug("ao audio device opened")
                else:
                    self.audiodev = aoaudiodev(self.aodevice, rate=self.rate, options=self.aooptions)
                    log.debug("ao audio device opened")
            except Exception, e:
                if not errorlogged:
                    log.debug_traceback()
                    log.error(_('cannot open audio device: error "%s"') % e)
                    errorlogged = True
                time.sleep(1)

    def closedevice(self):
        if self.audiodev is not None:
            # we use self.audiodev = None as a marker for a closed audio device
            # To avoid race conditions we thus first have to mark the device as
            # closed before really closing it. Note that when we try to reopen the device
            # later, and it has not yet been closed, the opendevice method will retry
            # this 
            openaudiodev = self.audiodev
            self.audiodev = None
            openaudiodev.close()
            log.debug("audio device closed")

    def queuelen(self):
        """ return length of currently buffered PCM data in seconds"""
        return 1.0/4.0*self.queue.qsize()*self.SIZE/self.rate

    def run(self):
        while not self.done:
            try:
                if self.ispaused:
                    self.restart.wait()
                    self.restart.clear()
                buff, bytes = self.queue.get(1)
                if buff != 0 and bytes != 0:
                    audiodev = self.audiodev
                    while audiodev is None:
                        self.opendevice()
                        audiodev = self.audiodev
                    audiodev.play(buff, bytes)
            except:
                log.warning("exception occured in bufferedaudiodev")
                log.debug_traceback()

    def play(self, buff, bytes):
        self.queue.put((buff, bytes))

    def flush(self):
        while True:
            try:
                self.queue.get(0)
            except Queue.Empty:
                break

    def pause(self):
        self.ispaused = True
        self.closedevice()

    def unpause(self):
        if self.ispaused:
            self.ispaused = False
            self.restart.set()

    def quit(self):
        self.done = True
        self.flush()
        self.closedevice()
        self.restart.set()

#
# wrapper class for playlistitems or songs
#

class decodedsong:
    def __init__(self, playlistitemorsong, rate, profiles):
        if isinstance(playlistitemorsong, services.playlist.playlistitem):
            self.song = playlistitemorsong.song
            self.playlistitem = playlistitemorsong
        else:
            self.song = playlistitemorsong
            self.playlistitem = None
        self.decodedsong = decoder.decodedsong(self.song, rate)
        self.replaygain = self.calculate_replaygain(["track"])

        # these method are handled by the decodedsong
        self.seekrelative = self.decodedsong.seekrelative
        self.playfaster = self.decodedsong.playfaster
        self.playslower = self.decodedsong.playslower
        self.resetplayspeed = self.decodedsong.resetplayspeed

    def __repr__(self):
        return "decodedsong(%r)" % repr(self.song)

    def read(self, size):
        # read decoded pcm stram and adjust for replaygain if necessary
        buff = self.decodedsong.read(size)
        if self.replaygain != 1:
            pcm.scale(buff, self.replaygain)
        return buff

    def succeedsonalbum(self, otherdecodedsong):
        " checks whether otherdeocedsong follows self on the same album "
        return (self.song.artist      and self.song.artist == otherdecodedsong.song.artist and
                self.song.album       and self.song.album == otherdecodedsong.song.album and
                self.song.tracknumber and otherdecodedsong.song.tracknumber and
                self.song.tracknumber == otherdecodedsong.song.tracknumber-1 )

    def calculate_replaygain(self, profiles):
       # the following code is adapted from quodlibet
       """Return the recommended Replay Gain scale factor.

       profiles is a list of Replay Gain profile names ('album',
       'track') to try before giving up. The special profile name
       'none' will cause no scaling to occur.
       """
       for profile in profiles:
           if profile is "none":
               return 1.0
           try:
               db = getattr(self.song, "replaygain_%s_gain" % profile)
               peak = getattr(self.song, "replaygain_%s_peak" % profile)
           except AttributeError:
               continue
           else:
               if db is not None and peak is not None:
                   scale = 10.**(db / 20)
                   if scale * peak > 1:
                       scale = 1.0 / peak # don't clip
                   return min(15, scale)
       else:
           return 1.0

    def rtime(self):
        " remaing playing time "
        return self.decodedsong.ttime - self.decodedsong.ptime

    def ptime(self):
        " playing time "
        return self.decodedsong.ptime



class player(genericplayer):

    def __init__(self, id, playlistid, autoplay, aodevice, aooptions, bufsize,
                 crossfading, crossfadingstart, crossfadingduration):
        self.rate = 44100
        self.SIZE = 4096
        self.volume = 1
        self._volume_scale = 0.005    # factor for logarthmic volume change

        # use C version of buffered audio device if present
        if bufferedao_present:
            self.audiodev = bufferedao.bufferedao(bufsize, self.SIZE, aodevice, byte_format=4, rate=self.rate, options=aooptions)
            # we have to start a new thread for the bufferedao device
            thread.start_new(self.audiodev.start, ())
            log.debug("bufferedao device opened")
        else:
            # create audio device thread
            self.audiodev = bufferedaudiodev(aodevice, aooptions, bufsize, self.rate, self.SIZE)
            self.audiodev.start()

        # songs currently playing
        self.decodedsongs = []

        self.crossfading = crossfading
        if self.crossfading:
            # crossfading parameters (all values in seconds and change/second, resp.)
            self.crossfadingduration = crossfadingduration
            self.crossfadingstart = crossfadingstart
            self.crossfadingratio = 0
            self.crossfadingrate = 1.0/self.rate/self.crossfadingduration

        # self.songtransitionmode determines the behaviour on transitions between two songs
        # possible values are "crossfade", "gapkill" and None (do nothing special)
        self.songtransitionmode = None

        genericplayer.__init__(self, id, playlistid, autoplay)

    def _flushqueue(self):
        """ delete internal player queue and flush audiodevice """
        self.decodedsongs = []
        self.audiodev.flush()
        self.audiodev.closedevice()

    def play(self):
        """decode songs and mix them together"""

        # unpause buffered ao if necessary
        self.audiodev.unpause()

        if len(self.decodedsongs) == 1:
            song = self.decodedsongs[0]
            buff = song.read(self.SIZE)
            if len(buff) > 0:
                if self.volume != 1:
                    pcm.scale(buff, self._volume_scale**(1-self.volume))
                self.audiodev.play(buff, len(buff))
            else:
                log.debug("internal player: song ends: %r (0 songs in queue)" % self.decodedsongs[0])
                del self.decodedsongs[0]

            # reset songtransition mode, but before possibly requesting a new song
            self.songtransitionmode = None

            if len(buff) == 0 or (self.crossfading and song.rtime() < self.crossfadingstart):
                self.requestnextsong()

        elif len(self.decodedsongs) == 2:
            if self.songtransitionmode == "crossfade":
                # perform crossfading
                buff1 = self.decodedsongs[0].read(self.SIZE)
                buff2 = self.decodedsongs[1].read(self.SIZE)

                if len(buff1) and len(buff2):
                    # normal operation: no song has ended
                    buff, self.crossfadingratio = pcm.mix(buff1, buff2,
                                                          self.crossfadingratio,
                                                          self.crossfadingrate)
                    if self.crossfadingratio >= 1:
                        self.crossfadingratio = 0
                        log.debug("internal player: song ends: %r (1 song in queue)" %
                              self.decodedsongs[0])
                        del self.decodedsongs[0]
                if len(buff1) == 0:
                    buff = buff2
                    self.crossfadingratio = 0
                    log.debug("internal player: song ends: %r (1 song in queue)" % self.decodedsongs[0])
                    del self.decodedsongs[0]
                if len(buff2) == 0:
                    buff = buff1
                    self.crossfadingratio = 0
                    log.debug("internal player: song ends: %r" % self.decodedsongs[-1])
                    del self.decodedsongs[-1]
                    log.debug("internal player: %d songs in queue" % len(self.decodedsongs))
            elif self.songtransitionmode == "gapkill":
                # just kill gap between songs
                buff = self.decodedsongs[0].read(self.SIZE)
                if len(buff) < self.SIZE:
                    log.debug("internal player: song ends: %r (1 song in queue)" % self.decodedsongs[0])
                    del self.decodedsongs[0]

                    buff2 = self.decodedsongs[0].read(self.SIZE)
                    if len(buff2) == 0:
                        log.debug("internal player: song ends: %r" % self.decodedsongs[0])
                        del self.decodedsongs[0]
                        log.debug("internal player: %d songs in queue" % len(self.decodedsongs))
                    else:
                        buff = buff + buff2
            else:
                # neither crossfading nor gap killing
                del self.decodedsongs[0]
                buff = self.decodedsongs[0].read(self.SIZE)

            if len(buff) > 0:
                if self.volume != 1:
                    pcm.scale(buff, self._volume_scale**(1-self.volume))
                self.audiodev.play(buff, len(buff))

        # update playbackinfo

        if len(self.decodedsongs) > 0:
            # determine which song is currently played
            # try to take the buffer length into account
            if self.songtransitionmode == "crossfade" and self.decodedsongs[-1].ptime()-self.audiodev.queuelen() >= 0:
                playingsong = self.decodedsongs[-1]
                self.playbackinfo.updatecrossfade(1)
            else:
                playingsong = self.decodedsongs[0]
                self.playbackinfo.updatecrossfade(0)

            time = int(max(playingsong.ptime()-self.audiodev.queuelen(), 0))

            self.playbackinfo.updatesong(playingsong.song)
            self.playbackinfo.updatetime(time)
        else:
            self.playbackinfo.stopped()

    def _playsong(self, song, manual):
        log.debug("internal player: new song: %r" % song)

        if self.ispaused():
            self._flushqueue()

        # we want maximally 2 songs in queue
        if len(self.decodedsongs) == 2:
            del self.decodedsongs[0]

        try:
            self.decodedsongs.append(decodedsong(song, self.rate, ["track"]))
            if self.crossfading:
                self.songtransitionmode = "crossfade"
                # Check whether two songs come after each other on an
                # album In such a case, we don't want to crossfade.  If,
                # however, the user has requested the song change (via
                # playerforward), we do want crossfading.
                if not manual and len(self.decodedsongs) == 2:
                    if self.decodedsongs[0].succeedsonalbum(self.decodedsongs[1]):
                        self.songtransitionmode = "gapkill"
                        log.debug("internal player: don't crossfade successive songs.")
        except (IOError, RuntimeError):
            log.warning(_('failed to open song "%r"') % str(song.url))

        log.debug("internal player: %d songs in queue" % len(self.decodedsongs))

    def _playerpause(self):
        self.audiodev.pause()

    def _playerstop(self):
        self._flushqueue()

    def _playerseekrelative(self, seconds):
        if len(self.decodedsongs) == 0:
            return
        if len(self.decodedsongs) == 2:
            # we refuse to seek backward during crossfading
            if seconds < 0:
                return
            del self.decodedsongs[0]
        song = self.decodedsongs[0]
        song.seekrelative(seconds)
        self.audiodev.flush()

    def _player_change_volume_relative(self, volume_adj):
        self.volume = max(0, min(1, self.volume + volume_adj/100.0))
        hub.notify(events.player_volume_changed(self.id, self.volume))

    def _playerplayfaster(self):
        if self.decodedsongs:
            self.decodedsongs[0].playfaster()

    def _playerplayslower(self):
        if self.decodedsongs:
            self.decodedsongs[0].playslower()

    def _playerspeedreset(self):
        if self.decodedsongs:
            self.decodedsongs[0].resetplayspeed()

    def _playerreleasedevice(self):
        self.audiodev.closedevice()

    def _playerquit(self):
        self.audiodev.quit()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2005 Jörg Lehmann <joerg@luga.de>

# Ogg Vorbis decoder interface by Byron Ellacott <bje@apnic.net>.
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path

import hub, requests
import log
import pcm
import encoding

#
# decoder class and simple decoder registry
#

class decoder:
    def __init__(self, path):
        pass

    def samplerate(self):
        """ return samplerate in samples per second """
        pass

    def ttime(self):
        """ return total length of song in seconds """
        pass

    def ptime(self):
        """ return the current position in the song in seconds """
        pass

    def read(self):
        """ return pcm stream (16bit, 2 channels) """
        pass

    def seekrelative(self, seconds):
        """ seek in stream by the given number of seconds (relative to current position) """
        pass


# mapping: file type -> decoder class
_decoders = {}

def registerdecoder(type, decoderclass):
    _decoders[type] = decoderclass

def getdecoder(type):
    return _decoders[type]

#
# MP3 decoder using libmad
#

class mp3decoder(decoder):
    def __init__(self, path):
        assert isinstance(path, str), "path has to be a string"
        self.file = mad.MadFile(path)

    def samplerate(self):
        return self.file.samplerate()

    def ttime(self):
        return self.file.total_time()//1000

    def ptime(self):
        return self.file.current_time()//1000

    def read(self):
        return self.file.read()

    def seekrelative(self, seconds):
        time = min(max(self.file.current_time() + seconds*1000, 0), self.file.total_time())
        self.file.seek_time(time)

try:
    import mad
    registerdecoder("mp3", mp3decoder)
except ImportError:
    pass

#
# Ogg Vorbis decoder
#

class oggvorbisdecoder(decoder):
    def __init__(self, path):
        self.file = ogg.vorbis.VorbisFile(path)

    def samplerate(self):
        return self.file.info().rate

    def ttime(self):
        return self.file.time_total(0)

    def ptime(self):
        return self.file.time_tell()

    def read(self):
        buff, bytes, bit = self.file.read()
        if self.file.info().channels == 2:
            return buffer(buff, 0, bytes)
        else:
            # for mono files, libvorbis really returns a mono stream
            # (as opposed to libmad) so that we have to "double" the
            # stream before we return it
            return pcm.upsample(buffer(buff, 0, bytes))

    def seekrelative(self, seconds):
        time = min(max(self.file.time_tell() + seconds, 0), self.file.time_total(0))
        self.file.time_seek(time)


try:
    import ogg.vorbis
    registerdecoder("ogg", oggvorbisdecoder)
except ImportError:
    pass

#
# FLAC decoder
#

class flacdecoder(decoder):
    def __init__(self, path):
        self.filedecoder = flac.decoder.FileDecoder()
        self.filedecoder.set_filename(path)
        # register callbacks
        self.filedecoder.set_write_callback(self._write_callback)
        self.filedecoder.set_error_callback(self._error_callback)
        self.filedecoder.set_metadata_callback(self._metadata_callback)
        # init decoder and process (here: ignore) metadata
        self.filedecoder.init()
        self.filedecoder.process_until_end_of_metadata()

        self._ptime = 0  # position in file in seconds

        # to be able to return the sample rate, we have to decode
        # some data
        self.buff = None
        self.filedecoder.process_single()

    def _metadata_callback(self, dec, block):
        if block.type == flac.metadata.STREAMINFO:
            streaminfo = block.data.stream_info
            self._samplerate = streaminfo.sample_rate
            self._channels = streaminfo.channels
            self._bits_per_sample = streaminfo.bits_per_sample
            self._ttime = streaminfo.total_samples // self._samplerate

    def _error_callback(self, dec, block):
        pass

    def _write_callback(self, dec, buff, size):
        self.buff = buff

    def samplerate(self):
        return self._samplerate

    def ttime(self):
        return self._ttime

    def ptime(self):
        return int(self._ptime)

    def read(self):
        if self.buff is None:
            self.filedecoder.process_single()
        if self.buff is not None:
            result = self.buff[:]
            self._ptime += 1.0*len(result)/self._channels/self._bits_per_sample*8/self._samplerate
            # ok, here it becomes very weird. There seems to be a problem with
            # the pyflac module, which does not occur (for me!) when
            # I insert the following code
            try:
                for i in range(100): pass
            except:
                pass
            self.buff = None
            return result

    def seekrelative(self, seconds):
        self._ptime += seconds
        self.filedecoder.seek_absolute(self._ptime * self._samplerate)

try:
    import flac.decoder
    import flac.metadata
    registerdecoder("flac", flacdecoder)
except ImportError:
    pass

#
# main class
#

class decodedsong:

    """ song decoder and rate converter

    This class is for decoding of a song and the conversion of the
    resulting pcm stream to a defined sample rate. Besides the
    constructor, there is only one method, namely read, which
    returns a pcm frame of or less than a given arbitrary size.

    """

    def __init__(self, song, outrate):
        self.outrate = outrate
        self.default_rate = outrate

        try:
            decoder = getdecoder(song.type)
        except:
            log.error("No decoder for song type '%r' registered "% song.type)
            raise RuntimeError("No decoder for song type '%r' registered "% song.type)

        url = encoding.encode_path(song.url)
        if url.startswith("file://"):
            dbstats = hub.request(requests.getdatabasestats(song.songdbid))
            if not dbstats.basedir:
                log.error("Currently only support for locally stored songs available")
                raise RuntimeError("Currently only support for locally stored songs available")
            path = os.path.join(dbstats.basedir, url[7:])
            self.decodedfile = decoder(path)
        else:
            log.error("Currently only support for locally stored songs available")
            raise RuntimeError("Currently only support for locally stored songs available")

        # Use the total time given by the decoder library and not the one
        # stored in the database. The former one turns out to be more precise
        # for some VBR songs.
        self.ttime = max(self.decodedfile.ttime(), song.length)

        # sometimes the mad library seems to report a wrong sample rate,
        # so use the one stored in the database
        if song.samplerate:
            self.samplerate = song.samplerate
        else:
            self.samplerate = self.decodedfile.samplerate()

        self.buff = self.last_l = self.last_r = None
        self.buffpos = 0
        self.ptime = 0

    def read(self, size):
        if self.buff is not None:
            bytesleft = len(self.buff) - self.buffpos
        else:
            bytesleft = 0

        # fill buffer, if necessary 
        while bytesleft < size:
            newbuff = self.decodedfile.read()
            if newbuff:
                self.buff, self.last_l, self.last_r = \
                           pcm.rate_convert(newbuff,
                                            self.samplerate,
                                            self.buff,
                                            self.buffpos,
                                            self.outrate,
                                            self.last_l,
                                            self.last_r)

                # the new self.buff contains only new data
                self.buffpos = 0
                bytesleft = len(self.buff)
            else:
                size = bytesleft
                break

        oldpos = self.buffpos
        self.buffpos += size
        self.ptime = self.decodedfile.ptime()
        if self.buff:
            return self.buff[oldpos:self.buffpos]
        else:
            return []

    def seekrelative(self, seconds):
        self.decodedfile.seekrelative(seconds)
        self.buff = self.last_l = self.last_r = None
        self.buffpos = 0
        self.ptime = self.decodedfile.ptime()

    def playslower(self, speed_adj = 441):
        self.outrate += speed_adj

    def playfaster(self, speed_adj = 441):
        # Its absurd that someone would try this
        # but we better check for it.
        if (self.outrate - speed_adj) < 1:
            self.outrate = 1
        else:
            self.outrate -= speed_adj

    def resetplayspeed(self):
        self.outrate = self.default_rate



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2006 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses, time

import config
import events, hub
import playlist
import statusbar
import window
import encoding

from helper import formattime

class playlistwin(window.window):
    def __init__(self,  screen, layout, channel, playerid):
        self.channel = channel
        self.keybindings = config.keybindings.playlistwindow
        self.songformat = config.playlistwindow.songformat
        h, w, y, x, border = layout
        window.window.__init__(self,
                               screen, h, w, y, x,
                               config.colors.playlistwindow,
                               _("Playlist"),
                               border, config.playlistwindow.scrollbar)

        # Immediately remove focus from playlist window in order to
        # prevent wrong selectionchanged events being issued, which
        # would lead to playing the first song in the playlist on the
        # second player (if configured)
        self.bottom()

        self.playlist = playlist.playlist(self, playerid)

        self.channel.subscribe(events.keypressed, self.keypressed)
        self.channel.subscribe(events.mouseevent, self.mouseevent)
        self.channel.subscribe(events.focuschanged, self.focuschanged)

    def updatestatusbar(self):
        sbar = []
        if self.playlist.selected is not None:
            sbar += statusbar.generatedescription("playlistwindow", "deleteitem")
            sbar += statusbar.separator
            sbar += statusbar.generatedescription("playlistwindow", "moveitemup")
            sbar += statusbar.separator
            sbar += statusbar.generatedescription("playlistwindow", "moveitemdown")
            sbar += statusbar.separator

        sbar += statusbar.generatedescription("playlistwindow", "activatefilelist")
        hub.notify(events.statusbar_update(0, sbar))

    def updatescrollbar(self):
        self.drawscrollbar(self.playlist.top, len(self.playlist))

    def resize(self, layout):
        h, w, y, x, self.border = layout
        window.window.resize(self, h, w, y, x)
        self.playlist._updatetop()
        if not self.hasfocus():
            self.playlist._recenter()

    def activatefilelist(self):
        # before recentering we remove the focus from the playlist in order
        # to prevent wrong songchanged events being issued (which would lead to
        # wrong songs being played on the secondary player)
        self.bottom()
        self.playlist._recenter()
        hub.notify(events.activatefilelist())

    # event handlers

    def keypressed(self, event):
        if self.hasfocus():
            key = event.key

            if key in self.keybindings["selectnext"]:
                self.playlist.selectnext()
            elif key in self.keybindings["selectprev"]:
                self.playlist.selectprev()
            elif key in self.keybindings["selectnextpage"]:
                self.playlist.selectnextpage()
            elif key in self.keybindings["selectprevpage"]:
                self.playlist.selectprevpage()
            elif key in self.keybindings["selectfirst"]:
                self.playlist.selectfirst()
            elif key in self.keybindings["selectlast"]:
                self.playlist.selectlast()
            elif key in self.keybindings["activatefilelist"]:
                self.activatefilelist()
            elif key in self.keybindings["moveitemup"]:
                self.playlist.moveitemup()
            elif key in self.keybindings["moveitemdown"]:
                self.playlist.moveitemdown()
            elif key in self.keybindings["deleteitem"]:
                self.playlist.deleteselected()
            elif key in self.keybindings["playselectedsong"]:
                self.playlist.playselected()
            elif key in self.keybindings["shuffle"]:
                hub.notify(events.playlistshuffle())
            elif key in self.keybindings["rescan"]:
                self.playlist.rescanselection(True)
            elif ord("0")<=key<=ord("5"):
                self.playlist.rateselection(key-ord("1")+1)
            elif key in self.keybindings["filelistjumptoselectedsong"]:
                self.playlist.filelistjumptoselected()
                self.activatefilelist()
            else:
                return

            self.update()
            raise hub.TerminateEventProcessing

    def mouseevent(self, event):
        if self.enclose(event.y, event.x):
            y, x = self.stdscrtowin(event.y, event.x)
            self.top()

            if event.state & curses.BUTTON1_CLICKED:
                if x==self.ix+self.iw and self.hasscrollbar:
                    scrollbarbegin, scrollbarheight = self.scrollbardimensions(self.playlist.top,
                                                                               len(self.playlist))
                    if y==self.iy+1:
                        self.playlist.selectprev()
                    elif y==self.iy+self.ih-2:
                        self.playlist.selectnext()
                    elif self.iy<y<scrollbarbegin:
                        self.playlist.selectprevpage()
                    elif scrollbarbegin+scrollbarheight<=y<self.iy+self.ih-2:
                        self.playlist.selectnextpage()
                else:
                    self.playlist.selectbylinenumber(y-1)
            elif event.state & curses.BUTTON1_DOUBLE_CLICKED:
                self.playlist.selectbylinenumber(y-1)
            elif event.state & curses.BUTTON3_CLICKED:
                pass
            else:
                return

            self.update()
            raise hub.TerminateEventProcessing

    def focuschanged(self, event):
        if self.hasfocus():
            hub.notify(events.selectionchanged(self.playlist.getselected()))
        self.update()

    # window update method

    def update(self):
        if self.playlist.autoplaymode == "repeat":
            autoplaymode = " [%s]" % _("Repeat")
        elif self.playlist.autoplaymode == "random":
            autoplaymode = " [%s]" % _("Random")
        else:
            autoplaymode = ""
        self.settitle("%s (-%s/%s)%s" % ( _("Playlist"),
                                        formattime((self.playlist.ttime-
                                                    self.playlist.ptime)),
                                        formattime(self.playlist.ttime),
                                        autoplaymode))

        window.window.update(self)
        if self.hasfocus():
            self.updatestatusbar()

        for i in range(self.playlist.top, self.playlist.top+self.ih):
            attr = curses.A_NORMAL

            if i<len(self.playlist):
                item = self.playlist[i]
                if item.playstarttime is not None:
                    h, m, s = time.localtime(item.playstarttime)[3:6]
                else:
                    h = m = s = 0
                adddict = {"playstarthours":   h,
                           "playstartminutes": m,
                           "playstartseconds": s}
                name = encoding.encode(item.song.format(self.songformat, adddict=adddict))
                if self.playlist.playingitem and item is self.playlist.playingitem:
                    if self.playlist.selected==i and self.hasfocus():
                        attr = self.colors.selected_playingsong
                    else:
                        attr = self.colors.playingsong
                elif item.hasbeenplayed():
                    if i==self.playlist.selected and self.hasfocus():
                        attr = self.colors.selected_playedsong
                    else:
                        attr = self.colors.playedsong
                else:
                    if i==self.playlist.selected and self.hasfocus():
                        attr = self.colors.selected_unplayedsong
                    else:
                        attr = self.colors.unplayedsong
            else:
                name = ""
            self.addnstr(i-self.playlist.top+self.iy, self.ix, name.ljust(self.iw)[:self.iw], self.iw, attr)

        self.updatescrollbar()

        # move cursor to the right position in order to make it more
        # easy for users of Braille displays to track the current
        # position/selection
	if self.hasfocus() and self.playlist.selected is not None:
	    self.win.move(self.playlist.selected-self.playlist.top+1, 1)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import item

class request:
    def __repr__(self):
        return self.__class__.__name__

    __repr__ = __repr__

#
# database requests
#

class dbrequest:
    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)

    def __cmp__(self, other):
        cmp(hash(self), hash(other))

    def __hash__(self):
        # for the cashing system every dbrequest has to be hashable
        # by default we rely on self.__repr__ for computing the hash value
        return hash(repr(self))

class dbrequestsingle(dbrequest):
    """ db request yielding a single result (not a list) and requiring a
    specific songdb to work on
    """
    pass

class dbrequestsongs(dbrequest):
    """ db request yielding a list of songs, which have to be merged when querying multiple databases

    Note that the resulting list must not be changed by the caller,
    with the exception that the order of the items may be changed at
    will (for instance by sorting)

    """

    def __init__(self, songdbid, random=False, sort=False, filters=None):
        self.songdbid = songdbid
        self.sort = sort
        self.random = random
        self.filters = filters

    def __repr__(self):
        return "%r(%r, %r, random=%r)->%r" % (self.__class__.__name__, self.sort, self.filters, self.random, self.songdbid)


class dbrequestlist(dbrequest):
    """ db request yielding a result list (not containing songs),
    which have to be merged when querying multiple databases

    Note that the resulting list must not be changed by the caller!
    """
    def __init__(self, songdbid, filters=None):
        self.songdbid = songdbid
        self.filters = filters

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.filters, self.songdbid)

#
# database requests which yield a single result
#

class getdatabasestats(dbrequest):
    """ return songdbstats instance for database """
    pass


class getsong_metadata(dbrequestsingle):
    """fetch song metadata from database songdbid corresponding to song_id""" 
    def __init__(self, songdbid, song_id):
        self.songdbid = songdbid
        self.song_id = song_id

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song_id, self.songdbid)


class gettag_id(dbrequestsingle):
    def __init__(self, songdbid, tag_name):
        self.songdbid = songdbid
        self.tag_name = tag_name

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.tag_name, self.songdbid)


#
# database requests which yield a list of songs
#

class getsongs(dbrequestsongs):
    pass


class getlastplayedsongs(dbrequestsongs):
    pass


#
# database requests which yield lists of other items
#

class getartists(dbrequestlist):
    pass


class getalbums(dbrequestlist):
    pass


class gettags(dbrequestlist):
    pass


class getratings(dbrequestlist):
    pass


class getplaylists(dbrequestlist):
    pass

#
# database request yielding the number of items of a certain kind
#

class getnumberofsongs(dbrequest):
    def __init__(self, songdbid, filters=None):
        self.songdbid = songdbid
        self.filters = filters

    def __repr__(self):
        return "%r(%r))->%r" % (self.__class__.__name__, self.filters, self.songdbid)


class dbrequestnumber(dbrequest):
    def __init__(self, songdbid, filters=None):
        self.songdbid = songdbid
        self.filters = filters

    def __repr__(self):
        return ( "%r(%r)->%r" % (self.__class__.__name__, self.filters, self.songdbid))


class getnumberofalbums(dbrequestnumber):
    pass

class getnumberofartists(dbrequestnumber):
    pass

class getnumberoftags(dbrequestnumber):
    pass

class getnumberofratings(dbrequestnumber):
    pass

# autoregisterer request

class autoregisterer_queryregistersong(dbrequest):
    def __init__(self, songdbid, path):
        self.songdbid = songdbid
        self.path = path

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.path, self.songdbid)

# songdbmanager

class getsongdbmanagerstats(request):
    """ request statistical information about songdbs and the request cache

    Returns services.songdb.songdbmanagerstats instance."""
    pass

#
# other requests for playlist and player service
#

class playlist_requestnextsong(request):
    """ request a playlistitem from playlistid. Go back in playlist if previous is set """
    def __init__(self, playlistid, previous=0):
        self.playlistid = playlistid
        self.previous = previous

    def __repr__(self):
        return "%r->%r,%r" % (self.__class__.__name__, self.playlistid, self.previous)


class getplaybackinfo(request):
    """ request info about song currently playing on player playerid """
    def __init__(self, playerid):
        self.playerid = playerid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, `self.playerid`)


class requestinput:
    def __init__(self, title, prompt, handler):
        self.title = title
        self.prompt = prompt
        self.handler = handler

    def __repr__(self):
        return "%r(%r,%r,%r)" % (self.__class__.__name__,
                              self.title, self.prompt, `self.handler`)


class playlistgetcontents(request):
    pass



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2004, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os, os.path, sys, getopt, locale

##############################################################################
# gettext initialization.
##############################################################################

# We have to do this very early, before importing our modules. We
# assume that the locales lie in same dir as this module. This may not
# be the case, if the .mo files are installed at their proper
# location.

try:
    import gettext
    locallocaledir = os.path.join(os.path.dirname(sys.argv[0]), "../locale")
    gettext.install("PyTone", locallocaledir)
except:
    # Disable localization if there is any problem with the above.
    # This works around a problem with Python 2.1
    import __builtin__
    __builtin__.__dict__['_'] = lambda s: s

import network, events, requests, version, helper

#
# parse command line options
#

server = None
port = 1972
unixsocketfile = None
debugmode = False

def usage():
    print "pytonectl %s" % version.version
    print "Copyright (C) 2003, 2004 Jörg Lehmann <joerg@luga.de>"
    print "usage: pytonectl.py [options] command"
    print
    print "Possible options are:"
    print "   -h, --help:              show this help"
    print "   -s, --server <hostname>: connect to PyTone server on hostname"
    print "   -p, --port <portnumber>: connect to PyTone server on given port"
    print "   -f, --file <filename>:   connect to PyTone UNIX socket filename"
    print "   -d, --debug:             enable debug mode"
    print
    print "The supported commands are:"
    print "    getplayerinfo:                  show information on the song currently being played"
    print "    playerforward:                  play the next song in the playlist"
    print "    playerprevious:                 play the previous song in the playlist"
    print "    playerseekrelative <seconds>:   seek relative in the current song by the given number of seconds"
    print "    playerpause:                    pause the player"
    print "    playerstart:                    start/unpause the player"
    print "    playertogglepause:              pause the player, if playing, or play, if paused"
    print "    playerstop:                     stop the player"
    print "    playerratecurrentsong <rating>: rate the song currently being played (1<=rating<=5)"
    print "    playlistaddsongs <filenames>:   add files to end of playlist"
    print "    playlistaddsongtop <filename>:  play file immediately"
    print "    playlistclear:                  clear the playlist"
    print "    playlistdeleteplayedsongs:      remove all played songs from the playlist"
    print "    playlistreplay:                 mark all songs in the playlist as unplayed"
    print "    playlistshuffle:                shuffle the playlist"

try:
    opts, args = getopt.getopt(sys.argv[1:],
                               "hs:p:f:d",
                               ["help", "server=", "port=", "file=", "debug"])
except getopt.GetoptError:
    usage()
    sys.exit(2)

for o, a in opts:
    if o in ("-h", "--help"):
        usage()
        sys.exit()
    if o in ("-s", "--server"):
        server = a
    if o in ("-p", "--port"):
        port = int(a)
    if o in ("-f", "--file"):
        unixsocketfile = a
    if o in ("-d", "--debug"):
        debugmode = True

# initialize the debug file if necessary
import log, sys
if debugmode:
    log.debugfile = sys.stdout
    log.info("Debug mode enabled")

if server is not None and unixsocketfile is not None:
    print "Error: cannot connect both via network and unix sockets"
    sys.exit(2)
if server is None:
    if unixsocketfile is None:
        unixsocketfile =  os.path.expanduser("~/.pytone/pytonectl")
    networklocation = unixsocketfile
else:
    networklocation = server, port

try:
    channel = network.clientchannel(networklocation)
except Exception, e:
    print "Error: cannot connect to PyTone server: %s" % e
    sys.exit(2)

channel.start()

if len(args)==0:
    usage()
    sys.exit(2)
elif len(args)==1:
    if args[0]=="playerforward":
        channel.notify(events.playernext("main"))
    elif args[0]=="playerprevious":
        channel.notify(events.playerprevious("main"))
    elif args[0]=="playerpause":
        channel.notify(events.playerpause("main"))
    elif args[0]=="playerstart":
        channel.notify(events.playerstart("main"))
    elif args[0]=="playertogglepause":
        channel.notify(events.playertogglepause("main"))
    elif args[0]=="playerstop":
        channel.notify(events.playerstop("main"))
    elif args[0]=="playlistclear":
        channel.notify(events.playlistclear())
    elif args[0]=="playlistdeleteplayedsongs":
        channel.notify(events.playlistdeleteplayedsongs())
    elif args[0]=="playlistreplay":
        channel.notify(events.playlistreplay())
    elif args[0]=="playlistshuffle":
        channel.notify(events.playlistshuffle())
    elif args[0]=="getplayerinfo":
        playbackinfo = channel.request(requests.getplaybackinfo("main"))
        if playbackinfo.song:
            # we have to manually request the song metadata because there the main event and request hub is not correctly
            # initialized
            song_metadata = channel.request(requests.getsong_metadata(playbackinfo.song.songdbid, playbackinfo.song.id))
           
            enc = locale.getpreferredencoding() or "ascii"

            s=u"%s - %s (%s/%s)" % (song_metadata.artist,
                                    song_metadata.title,
                                    helper.formattime(playbackinfo.time),
                                    helper.formattime(song_metadata.length))
            print s.encode(enc)
    else:
        usage()
        sys.exit(2)
else:
    if args[0]=="playerratecurrentsong" and len(args)==2:
        try:
            rating = int(args[1])
            if not 1<=rating<=5:
                raise
        except:
            usage()
            sys.exit(2)
        channel.notify(events.playerratecurrentsong("main", rating))
    if args[0]=="playerseekrelative" and len(args)==2:
        try:
            seconds = float(args[1])
        except:
            usage()
            sys.exit(2)
        channel.notify(events.playerseekrelative("main", seconds))
    elif args[0]=="playlistaddsongs":
        songs = [channel.request(requests.autoregisterer_queryregistersong("main", path)) for path in args[1:]]
        channel.notify(events.playlistaddsongs(songs))
    elif args[0]=="playlistaddsongtop" and len(args)==2:
        song = channel.request(requests.autoregisterer_queryregistersong("main", (args[1])))
        channel.notify(events.playlistaddsongtop(song))
    else:
        usage()
        sys.exit(2)

channel.quit()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import errors
import re
import config
import encoding
import events, hub
import log
import statusbar
import window

# two simple abstraction classes for the different mixer types from the oss
# and the ossaudiodev module, as well as the internal mixer

class ossmixer:
    def __init__(self, device, channel):
        self.mixer = oss.open_mixer(device)
        self.channel = channel
        log.info(_("initialized oss mixer: device %s, channel %s") %
                 (device, channel))

    def get(self):
        return self.mixer.read_channel(self.channel)

    def adjust(self, level_adjust):
        oldlevel = self.get()
        self.mixer.write_channel(self.channel, (max(0, min(oldlevel[0]+level_adjust, 100)),
                                                max(0, min(oldlevel[1]+level_adjust, 100)) ) )

class ossaudiodevmixer:
    def __init__(self, device, channel):
        self.mixer = oss.openmixer(device)
        self.channel = channel
        log.info(_("initialized oss mixer: device %s, channel %s") %
                 (device, channel))

    def get(self):
        return self.mixer.get(self.channel)


    def adjust(self, level_adjust):
        oldlevel = self.get()
        self.mixer.set(self.channel, (max(0, min(oldlevel[0]+level_adjust, 100)),
                                      max(0, min(oldlevel[1]+level_adjust, 100)) ) )

class internalmixer:
    def __init__(self, playerid):
        self.playerid = playerid
        self.volume = 1
        log.info(_("initialized internal mixer: player %s") % playerid)

    def get(self):
        return [self.volume*100, self.volume*100]

    def adjust(self, level_adjust):
        hub.notify(events.player_change_volume_relative(self.playerid, level_adjust))
        # to make the update smoother, we anticipate the volume change
        self.volume = max(0, min(1, self.volume + level_adjust/100.))



# determine oss module to be used, if any present

try:
    import ossaudiodev as oss
    externalmixer = ossaudiodevmixer 
except:
    try:
        import oss
        externalmixer = ossmixer
    except:
        externalmixer = None
    

class mixerwin(window.window):

    def __init__(self, screen, maxh, maxw, channel):
        self.channel = channel
        if config.mixer.type == "external":
            if externalmixer is not None:
                mixer_device = config.mixer.device
                channelre = re.compile("SOUND_MIXER_[a-zA-Z0-9]")
                if channelre.match(config.mixer.channel):
                    mixer_channel = eval("oss.%s" % config.mixer.channel)
                else:
                    raise errors.configurationerror("Wrong mixer channel specification: %s" % config.mixer.channel)
                self.mixer = externalmixer(mixer_device, mixer_channel)
            else:
                 self.mixer = internalmixer("main")
                 log.warning("Could not initialize external mixer, using internal one")
        elif config.mixer.type == "internal":
            self.mixer = internalmixer("main")
        else:
            self.mixer = None
        self.stepsize = config.mixer.stepsize

        if self.mixer:
            self.level = self.mixer.get()
        else:
            self.level = None
        self.keybindings = config.keybindings.general

        # for identification purposes, we only generate this once
        self.hidewindowevent = events.hidewindow(self)

        self.hide()

        if self.mixer:
            self.channel.subscribe(events.keypressed, self.keypressed)
            self.channel.subscribe(events.mouseevent, self.mouseevent)
            self.channel.subscribe(events.hidewindow, self.hidewindow)
            self.channel.subscribe(events.focuschanged, self.focuschanged)
        if isinstance(self.mixer, internalmixer):
            self.channel.subscribe(events.player_volume_changed, self.player_volume_changed)


    def changevolume(self, change):
        if self.mixer:
            self.mixer.adjust(change)

    # event handler

    def keypressed(self, event):
        key = event.key
        if key in self.keybindings["volumeup"]:
            self.changevolume(self.stepsize)
        elif key in self.keybindings["volumedown"]:
            self.changevolume(-self.stepsize)
        else:
            if self.hasfocus():
                self.hide()
                raise hub.TerminateEventProcessing
            return

        self.update()
        if config.mixerwindow.autoclosetime:
            hub.notify(events.sendeventin(self.hidewindowevent,
                                              config.mixerwindow.autoclosetime,
                                              replace=1))
        
        raise hub.TerminateEventProcessing

    def mouseevent(self, event):
        if self.hasfocus():
            self.hide()
            raise hub.TerminateEventProcessing

    def focuschanged(self, event):
        if self.hasfocus():
            sbar = statusbar.generatedescription("general", "volumedown")
            sbar += statusbar.separator
            sbar += statusbar.generatedescription("general", "volumeup")
            sbar += statusbar.terminate

            hub.notify(events.statusbar_update(0, sbar))

    def player_volume_changed(self, event):
        if isinstance(self.mixer, internalmixer) and event.playerid==self.mixer.playerid:
            self.mixer.volume = event.volume
            if self.hasfocus():
                self.update()

    def update(self):
        self.level = self.mixer.get()
        self.top()
        window.window.update(self)
        self.addstr(self.iy, self.ix, encoding.encode(_("Volume:")), self.colors.description)
        self.addstr(" %3d " % round(self.level[0]), self.colors.content)
 
        percent = int(round(self.barlen*self.level[0]/100))
        self.addstr("#"*percent, self.colors.barhigh)
        self.addstr("#"*(self.barlen-percent), self.colors.bar)


class popupmixerwin(mixerwin):

    """ mixer which appears as a popup at the center of the screen """

    def __init__(self, screen, maxh, maxw, channel):
        # calculate size and position
        self.barlen = 20
        h = 3
        w = len(_("Volume:")) + 7 + self.barlen
        y = (maxh-h)/2
        x = (maxw-w)/2

        window.window.__init__(self,
                               screen, h, w, y, x,
                               config.colors.mixerwindow,
                               _("Mixer"))

        mixerwin.__init__(self, screen, maxh, maxw, channel)

    def resize(self, maxh, maxw):
        h = 3
        w = len(_("Volume:")) + 7 + self.barlen
        y = (maxh-h)/2
        x = (maxw-w)/2
        window.window.resize(self, h, w, y, x)


class statusbarmixerwin(mixerwin):

    """ mixer which appears in the statusbar """

    def __init__(self, screen, maxh, maxw, channel):
        # calculate size and position
        self.barlen = max(0, maxw - len(_("Volume:")) - 5)

        window.window.__init__(self,
                               screen, 1, maxw, maxh-1, 0, config.colors.mixerwindow)

        mixerwin.__init__(self, screen, maxh, maxw, channel)

    def resize(self, maxh, maxw):
        window.window.resize(self, 1, maxw, maxh-1, 0)
